"""
This file contains a factory implementation for the controllers.
This is used to manage the imports and reduce effort in adding new controllers.
"""

import os
from importlib import import_module
from inspect import getmembers, isabstract, isclass

from fitbenchmarking.controllers.base_controller import Controller
from fitbenchmarking.utils.exceptions import (MissingSoftwareError,
                                              NoControllerError)


class ControllerFactory:
    """
    A factory for creating software controllers.
    This has the capability to select the correct controller, import it, and
    generate an instance of it.
    Controllers generated from this must be a subclass of
    base_controller.Controller
    """

    @staticmethod
    def create_controller(software):
        """
        Create a controller that matches the required software.

        :param software: The name of the software to create a controller for
        :type software: string

        :return: Controller class for the problem
        :rtype: fitbenchmarking.fitting.base_controller.Controller subclass
        """

        module_name = '{}_controller'.format(software.lower())

        try:
            module = import_module('.' + module_name, __package__)
        except ImportError as e:
            full_path = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                     module_name+'.py'))
            if os.path.exists(full_path):
                raise MissingSoftwareError('Requirements are missing for the '
                                           '{} controller: {}'.format(
                                               software, e))
            else:
                raise NoControllerError('Could not find controller for {}. '
                                        'Check the input is correct and try '
                                        'again.'.format(software))

        classes = getmembers(module, lambda m: (isclass(m)
                                                and not isabstract(m)
                                                and issubclass(m, Controller)
                                                and m is not Controller))

        return classes[0][1]



'''
This file will handle all interaction with the options configuration file.
'''

import configparser

import os
import numpy as np

from fitbenchmarking.utils.exceptions import OptionsError


class Options(object):
    """
    An options class to store and handle all options for fitbenchmarking
    """

    DEFAULTS = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                            'default_options.ini'))
    VALID_SECTIONS = ['MINIMIZERS', 'FITTING', 'PLOTTING', 'LOGGING']
    VALID_MINIMIZERS = \
        {'bumps': ['amoeba', 'lm-bumps', 'newton', 'de', 'mp'],
         'dfo': ['dfogn', 'dfols'],
         'gsl': ['lmsder', 'lmder', 'nmsimplex', 'nmsimplex2',
                 'conjugate_pr', 'conjugate_fr', 'vector_bfgs',
                 'vector_bfgs2', 'steepest_descent'],
         'mantid': ['BFGS',
                    'Conjugate gradient (Fletcher-Reeves imp.)',
                    'Conjugate gradient (Polak-Ribiere imp.)',
                    'Damped GaussNewton', 'Levenberg-Marquardt',
                    'Levenberg-MarquardtMD', 'Simplex',
                    'SteepestDescent', 'Trust Region'],
         'minuit': ['minuit'],
         'ralfit': ['gn', 'gn_reg', 'hybrid', 'hybrid_reg'],
         'scipy': ['Nelder-Mead', 'Powell', 'CG', 'BFGS',
                   'Newton-CG', 'L-BFGS-B', 'TNC', 'SLSQP'],
         'scipy_ls': ['lm-scipy-no-jac', 'lm-scipy', 'trf',
                      'dogbox']}
    VALID_FITTING = \
        {'algorithm_type': ['all', 'ls', 'deriv_free', 'general'],
         'software': ['bumps', 'dfo', 'gsl', 'mantid', 'minuit',
                      'ralfit', 'scipy', 'scipy_ls'],
         'use_errors': [True, False],
         'jac_method': ['SciPyFD'],
         'num_method': ['2point', '3point', 'cs']}
    VALID_PLOTTING = \
        {'make_plots': [True, False],
         'comparison_mode': ['abs', 'rel', 'both'],
         'table_type': ['acc', 'runtime', 'compare', 'local_min']}
    VALID_LOGGING = \
        {'level': ['NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR',
                   'CRITICAL'],
         'append': [True, False],
         'external_output': ['debug','display','log_only']}

    VALID = {'MINIMIZERS': VALID_MINIMIZERS,
             'FITTING': VALID_FITTING,
             'PLOTTING': VALID_PLOTTING,
             'LOGGING': VALID_LOGGING}

    DEFAULT_MINIMZERS = \
        {'bumps': ['amoeba', 'lm-bumps', 'newton', 'mp'],
         'dfo': ['dfogn', 'dfols'],
         'gsl': ['lmsder', 'lmder', 'nmsimplex', 'nmsimplex2',
                 'conjugate_pr', 'conjugate_fr', 'vector_bfgs',
                 'vector_bfgs2', 'steepest_descent'],
         'mantid': ['BFGS',
                    'Conjugate gradient (Fletcher-Reeves imp.)',
                    'Conjugate gradient (Polak-Ribiere imp.)',
                    'Damped GaussNewton', 'Levenberg-Marquardt',
                    'Levenberg-MarquardtMD', 'Simplex',
                    'SteepestDescent', 'Trust Region'],
         'minuit': ['minuit'],
         'ralfit': ['gn', 'gn_reg', 'hybrid', 'hybrid_reg'],
         'scipy': ['Nelder-Mead', 'Powell', 'CG', 'BFGS',
                   'Newton-CG', 'L-BFGS-B', 'TNC'],
         'scipy_ls': ['lm-scipy-no-jac', 'lm-scipy', 'trf',
                      'dogbox']}
    DEFAULT_FITTING = \
        {'num_runs': 5,
         'algorithm_type': 'all',
         'software': ['scipy', 'scipy_ls'],
         'use_errors': True,
         'jac_method': 'SciPyFD',
         'num_method': '2point'}
    DEFAULT_PLOTTING = \
        {'make_plots': True,
         'colour_scale': [(1.1, "#fef0d9"),
                          (1.33, "#fdcc8a"),
                          (1.75, "#fc8d59"),
                          (3, "#e34a33"),
                          (np.inf, "#b30000")],
         'comparison_mode': 'both',
         'table_type': ['acc', 'runtime', 'compare', 'local_min'],
         'results_dir': 'fitbenchmarking_results'}
    DEFAULT_LOGGING = \
        {'file_name': 'fitbenchmarking.log',
         'append': False,
         'level': 'INFO',
         'external_output': 'log_only'}
    DEFAULTS = {'MINIMIZERS': DEFAULT_MINIMZERS,
                'FITTING': DEFAULT_FITTING,
                'PLOTTING': DEFAULT_PLOTTING,
                'LOGGING': DEFAULT_LOGGING}

    def __init__(self, file_name=None):
        """
        Initialise the options from a file if file is given.
        Priority is values in the file, failing that, values are taken from
        DEFAULTS (stored in ./default_options.ini)

        :param file_name: The options file to load
        :type file_name: str
        """
        # stores the file name to be used to reset options for multiple
        # problem groups
        self.stored_file_name = file_name
        self.error_message = []
        self._results_dir = ''
        config = configparser.ConfigParser(converters={'list': read_list,
                                                       'str': str},
                                           allow_no_value=True)

        for section in self.VALID_SECTIONS:
            config.add_section(section)

        if file_name is not None:
            config.read(file_name)

            # Checks that the user defined sections are valid
            if config.sections() != self.VALID_SECTIONS:
                raise OptionsError(
                    "Invalid options sections set, {0}, the valid sections "
                    "are {1}".format(config.sections(), self.VALID_SECTIONS))
            config.sections()

            # Checks that the options within the sections are valid
            for key in self.VALID_SECTIONS:
                default_options_list = list(self.DEFAULTS[key].keys())
                user_options_list = [option[0] for option in config.items(key)]
                if not (set(user_options_list) <= set(default_options_list)):
                    raise OptionsError(
                        "Invalid options key set: \n{0}, \n the valid keys "
                        "are: \n{1}".format(user_options_list,
                                            default_options_list))

        minimizers = config['MINIMIZERS']
        self._minimizers = {}
        for key in self.VALID_FITTING["software"]:
            self._minimizers[key] = self.read_value(minimizers.getlist,
                                                    key)

        fitting = config['FITTING']
        self.num_runs = self.read_value(fitting.getint, 'num_runs')
        self.algorithm_type = self.read_value(fitting.getstr, 'algorithm_type')
        self.software = self.read_value(fitting.getlist, 'software')
        self.use_errors = self.read_value(fitting.getboolean, 'use_errors')
        self.jac_method = self.read_value(fitting.getstr, 'jac_method')
        self.num_method = self.read_value(fitting.getstr, 'num_method')

        plotting = config['PLOTTING']
        self.make_plots = self.read_value(plotting.getboolean, 'make_plots')
        self.colour_scale = self.read_value(plotting.getlist, 'colour_scale')
        check = [isinstance(c, tuple) for c in self.colour_scale]
        if check.count(False) == len(check):
            self.colour_scale = [(float(cs.split(',', 1)[0].strip()),
                                  cs.split(',', 1)[1].strip())
                                 for cs in self.colour_scale]
        self.comparison_mode = self.read_value(plotting.getstr,
                                               'comparison_mode')
        self.table_type = self.read_value(plotting.getlist, 'table_type')
        self.results_dir = self.read_value(plotting.getstr, 'results_dir')

        logging = config['LOGGING']
        self.log_append = self.read_value(logging.getboolean, 'append')
        self.log_file = self.read_value(logging.getstr, 'file_name')
        self.log_level = self.read_value(logging.getstr, 'level')

        self.external_output = self.read_value(logging.getstr,
                                               'external_output')

        if self.error_message != []:
            raise OptionsError('\n'.join(self.error_message))

    def reset(self):
        """
        Resets options object when running multiple problem groups.
        """
        self.__init__(self.stored_file_name)

    def read_value(self, func, option):
        """
        Helper function which loads in the value

        :param func: configparser function
        :type func: callable
        :param option: option to be read for file
        :type option: str

        :return: value of the option
        :rtype: list/str/int/bool
        """
        section = func.__str__().split("Section: ")[1].split('>')[0]
        try:
            value = func(option, fallback=self.DEFAULTS[section][option])
        except ValueError:
            self.error_message.append(
                "Incorrect options type for {}".format(option))
            value = None

        if option in self.VALID[section]:
            if isinstance(value, list):
                set1 = set(value)
                set2 = set(self.VALID[section][option])
                value_check = set1.issubset(set2)
            else:
                value_check = value in self.VALID[section][option]
            if not value_check:
                self.error_message.append(
                    "The option '{0}: {1}' in the ini file is invalid. {0} "
                    "must be on or more of {2}".format(
                        option, value, self.VALID[section][option]))
        return value

    @property
    def results_dir(self):
        return self._results_dir

    @results_dir.setter
    def results_dir(self, value):
        self._results_dir = os.path.abspath(value)

    @property
    def minimizers(self):
        return {s: self._minimizers[s] for s in self.software}

    @minimizers.setter
    def minimizers(self, value):
        self._minimizers = value

    def _create_config(self):
        """
        Return the contents of the options object as a ConfigParser object,
        which e.g. then can be written to a file object or other stream

        :return: ConfigParser
        """
        config = configparser.ConfigParser(converters={'list': read_list,
                                                       'str': str})

        def list_to_string(l):
            return '\n'.join(l)

        config['MINIMIZERS'] = {k: list_to_string(m)
                                for k, m in self.minimizers.items()}
        config['FITTING'] = {'num_runs': self.num_runs,
                             'algorithm_type': self.algorithm_type,
                             'software': list_to_string(self.software),
                             'use_errors': self.use_errors,
                             'jac_method': self.jac_method,
                             'num_method': self.num_method}
        cs = list_to_string(['{0}, {1}'.format(*pair)
                             for pair in self.colour_scale])
        config['PLOTTING'] = {'colour_scale': cs,
                              'comparison_mode': self.comparison_mode,
                              'make_plots': self.make_plots,
                              'results_dir': self.results_dir,
                              'table_type': list_to_string(self.table_type)}
        config['LOGGING'] = {'file_name': self.log_file,
                             'level': self.log_level,
                             'append': self.log_append,
                             'external_output': self.external_output}

        return config

    def write(self, file_name):
        """
        Write the contents of the options object to a new options file.

        :param file_name: The path to the new options file
        :type file_name: str
        """
        config = self._create_config()

        with open(file_name, 'w') as f:
            config.write(f)

    def write_to_stream(self, file_object):
        """
        Write the contents of the options object to a file object.

        :type file_object: file object
        """
        self._create_config().write(file_object)


def read_list(s):
    """
    Utility function to allow lists to be read by the config parser

    :param s: string to convert to a list
    :type s: string

    :return: list of items
    :rtype: list of str
    """
    return str(s).split('\n')



"""
This file implements a parser for the Fitbenchmark data format.
"""

from __future__ import absolute_import, division, print_function

import os
from collections import OrderedDict

import numpy as np

from fitbenchmarking.parsing.base_parser import Parser
from fitbenchmarking.parsing.fitting_problem import FittingProblem
from fitbenchmarking.utils.exceptions import MissingSoftwareError, ParsingError
from fitbenchmarking.utils.log import get_logger

LOGGER = get_logger()

import_success = {}
try:
    import mantid.simpleapi as msapi
    import_success['mantid'] = (True, None)
except ImportError as ex:
    import_success['mantid'] = (False, ex)


try:
    from sasmodels.data import empty_data1D
    from sasmodels.core import load_model
    from sasmodels.bumps_model import Experiment, Model
    import_success['sasview'] = (True, None)
except ImportError as ex:
    import_success['sasview'] = (False, ex)


# By design the parsers may require many the private methods
# pylint: disable=too-many-branches
class FitbenchmarkParser(Parser):
    """
    Parser for the native FitBenchmarking problem definition (FitBenchmark)
    file.
    """

    def __init__(self, filename, options):
        super(FitbenchmarkParser, self).__init__(filename, options)

        self._entries = None
        self._parsed_func = None

    def parse(self):
        """
        Parse the Fitbenchmark problem file into a Fitting Problem.

        :return: The fully parsed fitting problem
        :rtype: fitbenchmarking.parsing.fitting_problem.FittingProblem
        """
        # pylint: disable=attribute-defined-outside-init
        fitting_problem = FittingProblem(self.options)

        self._entries = self._get_data_problem_entries()
        software = self._entries['software'].lower()
        if not (software in import_success and import_success[software][0]):
            error = import_success[software][1]
            raise MissingSoftwareError('Requirements are missing for {} parser'
                                       ': {}'.format(software, error))

        self._parsed_func = self._parse_function()

        if software == 'mantid' and self._entries['input_file'][0] == '[':
            fitting_problem.multifit = True

        # NAME
        fitting_problem.name = self._entries['name']

        # DATA
        data_points = [_get_data_points(p) for p in self._get_data_file()]

        # FUNCTION
        if software == 'mantid':
            fitting_problem.function = self._create_mantid_function()
        elif software == 'sasview':
            fitting_problem.function = self._create_sasview_function()

        # EQUATION
        equation_count = len(self._parsed_func)
        if equation_count == 1:
            fitting_problem.equation = self._parsed_func[0]['name']
        else:
            fitting_problem.equation = '{} Functions'.format(equation_count)

        # STARTING VALUES
        fitting_problem.starting_values = self._get_starting_values()

        # PARAMETER RANGES
        vr = _parse_range(self._entries.get('parameter_ranges', ''))
        fitting_problem.value_ranges = vr if vr != {} else None

        # FIT RANGES
        fit_ranges_str = self._entries.get('fit_ranges', '')
        # this creates a list of strs like '{key: val, ...}' and parses each
        fit_ranges = [_parse_range('{' + r.split('}')[0] + '}')
                      for r in fit_ranges_str.split('{')[1:]]

        if fitting_problem.multifit:
            num_files = len(data_points)
            fitting_problem.data_x = [d[:, 0] for d in data_points]
            fitting_problem.data_y = [d[:, 1] for d in data_points]
            fitting_problem.data_e = [d[:, 2] if d.shape[1] > 2 else None
                                      for d in data_points]

            if not fit_ranges:
                fit_ranges = [{} for _ in range(num_files)]

            fitting_problem.start_x = [f['x'][0] if 'x' in f else None
                                       for f in fit_ranges]
            fitting_problem.end_x = [f['x'][1] if 'x' in f else None
                                     for f in fit_ranges]

        else:
            fitting_problem.data_x = data_points[0][:, 0]
            fitting_problem.data_y = data_points[0][:, 1]
            if data_points[0].shape[1] > 2:
                fitting_problem.data_e = data_points[0][:, 2]

            if fit_ranges and 'x' in fit_ranges[0]:
                fitting_problem.start_x = fit_ranges[0]['x'][0]
                fitting_problem.end_x = fit_ranges[0]['x'][1]

        if software == 'mantid':
            # String containing the function name(s) and the starting parameter
            # values for each function.
            fitting_problem.additional_info['mantid_equation'] \
                = self._entries['function']

            if fitting_problem.multifit:
                fitting_problem.additional_info['mantid_ties'] \
                    = self._parse_ties()

        return fitting_problem

    def _get_data_file(self):
        """
        Find/create the (full) path to a data_file specified in a FitBenchmark
        definition file, where the data_file is searched for in the directory
        of the definition file and subfolders of this file.

        :return: (full) path to a data file. Return None if not found
        :rtype: str or None
        """
        data_file_name = self._entries['input_file']
        if data_file_name.startswith('['):
            # Parse list assuming filenames do not have quote symbols or commas
            data_file_names = [
                d.replace('"', '').replace("'", '').strip('[').strip(']')
                for d in data_file_name.split(',')]
        else:
            data_file_names = [data_file_name]

        # find or search for path for data_file_name
        data_files = []
        for data_file_name in data_file_names:
            data_file = None
            for root, _, files in os.walk(os.path.dirname(self._filename)):
                for name in files:
                    if data_file_name == name:
                        data_file = os.path.join(root, data_file_name)

            if data_file is None:
                LOGGER.error("Data file %s not found", data_file_name)

            data_files.append(data_file)

        return data_files

    def _get_data_problem_entries(self):
        """
        Get the problem entries from a problem definition file.

        :return: The entries from the file with string values
        :rtype: dict
        """

        entries = {}
        for line in self.file.readlines():
            # Discard comments
            line = line.split('#', 1)[0]
            if line.strip() == '':
                continue

            lhs, rhs = line.split("=", 1)
            entries[lhs.strip()] = rhs.strip().strip('"').strip("'")

        return entries

    def _parse_function(self):
        """
        Get the params from the function as a list of dicts from the data
        file.

        :return: Function definition in format:
                 [{name1: value1, name2: value2, ...}, ...]
        :rtype: list of dict
        """
        # pylint: disable=too-many-branches, too-many-statements
        function_def = []

        for f in self._entries['function'].split(';'):
            params_dict = OrderedDict()
            pop_stack = 0

            stack = [params_dict]
            for p in f.split(','):
                name, val = p.split('=', 1)
                name = name.strip()
                val = val.strip()

                l_count = val.count('(')
                r_count = val.count(')')
                if l_count > r_count:
                    # in brackets
                    # should be of the form 'varname=(othervar=3, ...)'

                    # Allow for nested brackets e.g. 'a=(b=(c=(d=1,e=2)))'
                    for _ in range(l_count - r_count):
                        # Cover case where formula mistyped
                        if '=' not in val:
                            raise ParsingError('Unbalanced brackets in '
                                               'function value: {}'.format(p))
                        # Must start with brackets
                        if val[0] != '(':
                            raise ParsingError('Bad placement of opening '
                                               'bracket in function: '
                                               '{}'.format(p))
                        # Create new dict for this entry and put at top of
                        # working stack
                        new_dict = OrderedDict()
                        stack[-1][name] = new_dict
                        stack.append(new_dict)
                        # Update name and val
                        name, val = val[1:].split('=', 1)
                elif l_count == r_count:
                    # Check if single item in brackets
                    while '=' in val:
                        if val[0] == '(' and val[-1] == ')':
                            val = val[1:-1]
                            new_dict = OrderedDict()
                            stack[-1][name] = new_dict
                            stack.append(new_dict)
                            name, val = val.split('=', 1)
                            pop_stack += 1
                        else:
                            raise ParsingError('Function value contains '
                                               'unexpected "=": {}'.format(p))
                elif l_count < r_count:
                    # exiting brackets
                    pop_stack = r_count - l_count
                    # must end with brackets
                    if val[-pop_stack:] != ')' * pop_stack:
                        raise ParsingError('Bad placement of closing bracket '
                                           'in function: {}'.format(p))
                    val = val[:-pop_stack]

                # Parse to an int/float if possible else assume string
                tmp_val = None
                for t in [int, float]:
                    if tmp_val is None:
                        try:
                            tmp_val = t(val)
                        except (ValueError, TypeError):
                            pass

                if tmp_val is not None:
                    val = tmp_val

                stack[-1][name] = val

                if pop_stack > 0:
                    if len(stack) <= pop_stack:
                        raise ParsingError('Too many closing brackets in '
                                           'function: {}'.format(p))
                    stack = stack[:-pop_stack]
                    pop_stack = 0

            if len(stack) != 1:
                raise ParsingError('Not all brackets are closed in function.')
            function_def.append(params_dict)

        return function_def

    def _get_starting_values(self):
        """
        Get the starting values for the problem

        :return: Starting values for the function
        :rtype: list of OrderedDict
        """
        # Mantid functions can have reserved keywords so ignore these as they
        # are not parameters.
        ignore = ['name', 'BinWidth', 'ties', 'Formula', 'constraints']

        name_template = '{1}' if len(self._parsed_func) == 1 else 'f{0}_{1}'
        starting_values = [
            OrderedDict([(name_template.format(i, name), val)
                         for i, f in enumerate(self._parsed_func)
                         for name, val in f.items()
                         if name not in ignore])]

        return starting_values

    def _create_mantid_function(self):
        """
        Processing the function in the FitBenchmark problem definition into a
        python callable.

        :return: A callable function
        :rtype: callable
        """
        fit_function = None

        for f in self._parsed_func:
            name = f['name']
            params = f.copy()
            for key in ['name', 'ties']:
                if key in params:
                    params.pop(key)
            tmp_function = msapi.__dict__[name](**params)
            if 'ties' in f:
                tmp_function.tie(f['ties'])
            if fit_function is None:
                fit_function = tmp_function
            else:
                fit_function += tmp_function

        return fit_function

    def _create_sasview_function(self):
        """
        Creates callable function

        :return: the model
        :rtype: callable
        """
        equation = self._parsed_func[0]['name']
        starting_values = self._get_starting_values()
        value_ranges = _parse_range(self._entries.get('parameter_ranges', ''))
        param_names = starting_values[0].keys()

        def fitFunction(x, *tmp_params):

            model = load_model(equation)

            data = empty_data1D(x)
            param_dict = dict(zip(param_names, tmp_params))

            model_wrapper = Model(model, **param_dict)
            if value_ranges is not None:
                for name, values in value_ranges.items():
                    model_wrapper.__dict__[name].range(values[0], values[1])
            func_wrapper = Experiment(data=data, model=model_wrapper)

            return func_wrapper.theory()

        return fitFunction

    def _parse_ties(self):
        try:
            ties = []
            for t in self._entries['ties'].split(','):
                # Strip out these chars
                for s in '[] "\'':
                    t = t.replace(s, '')
                ties.append(t)

        except KeyError:
            ties = []
        return ties


def _parse_range(range_str):
    """
    Parse a range string for the problem into a dict or list of dict if
    multi-fit.

    :param range_str: The a string to parse
    :type range_str: string

    :return: The ranges in a dictionary with key as the var and value as a
             list with min and max
             e.g. {'x': [0, 10]}
    :rtype: dict
    """
    if not range_str:
        return {}

    output_ranges = {}
    range_str = range_str.strip('{').strip('}')
    tmp_ranges = range_str.split(',')
    ranges = []
    cur_str = ''
    for r in tmp_ranges:
        cur_str += r
        balanced = True
        for lb, rb in ['[]', '{}', '()']:
            if cur_str.count(lb) > cur_str.count(rb):
                balanced = False
            elif cur_str.count(lb) < cur_str.count(rb):
                raise ParsingError(
                    'Unbalanced brackets in range: {}'.format(r))
        if balanced:
            ranges.append(cur_str)
            cur_str = ''
        else:
            cur_str += ','

    for r in ranges:
        name, val = r.split(':')
        name = name.strip().strip('"').strip("'").lower()

        # Strip off brackets and split on comma
        val = val.strip(' ')[1:-1].split(',')
        val = [v.strip() for v in val]
        try:
            pair = [float(val[0]), float(val[1])]
        except ValueError:
            raise ParsingError('Expected floats in range: {}'.format(r))

        if pair[0] >= pair[1]:
            raise ParsingError('Min value must be smaller than max value '
                               'in range: {}'.format(r))

        output_ranges[name] = pair

    return output_ranges


def _get_data_points(data_file_path):
    """
    Get the data points of the problem from the data file.

    :param data_file_path: The path to the file to load the points from
    :type data_file_path: str

    :return: data points
    :rtype: np.ndarray
    """

    with open(data_file_path, 'r') as f:
        data_text = f.readlines()

    # Find the line where data starts
    # i.e. the first line with a float on it
    first_row = 0

    # Loop until break statement
    while True:
        try:
            line = data_text[first_row].strip()
        except IndexError:
            raise ParsingError('Could not find data points')
        if line != '':
            x_val = line.split()[0]
            try:
                _ = float(x_val)
            except ValueError:
                pass
            else:
                break
        first_row += 1

    dim = len(data_text[first_row].split())
    data_points = np.zeros((len(data_text) - first_row, dim))

    for idx, line in enumerate(data_text[first_row:]):
        point_text = line.split()
        # Skip any values that can't be represented
        try:
            point = [float(val) for val in point_text]
        except ValueError:
            point = [np.nan for val in point_text]
        data_points[idx, :] = point

    # Strip all np.nan entries
    data_points = data_points[~np.isnan(data_points[:, 0]), :]

    return data_points



"""
This script was used to extract x, y, and e data from Mantid nexus files.
This relies on having Mantid installed and is only intended for use by
developers of FitBenchmarking.
"""

import itertools
import os

from mantid import simpleapi as ms


def parse_nxs(file_name):
    name, ext = os.path.splitext(file_name)
    if ext != '.nxs':
        print('Wrong file!', ext)
        return -1

    ws = ms.LoadNexus(file_name)
    xs = ws.readX(0)
    ys = ws.readY(0)
    es = ws.readE(0)

    data_template = '{0} {1} {2}\n'

    lines = ['# X Y E\n', '\n']
    lines += [data_template.format(x, y, e)
              for x, y, e in itertools.izip(xs, ys, es)]

    with open(name + '.txt', 'w') as f:
        f.writelines(lines)



"""
Functions that prepare the function specified in NIST problem definition file
into the right format for SciPy.
"""

from __future__ import (absolute_import, division, print_function)

# This import is needed for dynamic scipy function def
import numpy as np

from fitbenchmarking.utils.exceptions import ParsingError


def nist_func_definition(function, param_names):
    """
    Processing a function plus different set of starting values as specified in
    the NIST problem definition file into a callable

    :param function: function string as defined in a NIST problem definition
                     file
    :type function: str
    :param param_names: names of the parameters in the function
    :type param_names: list

    :return: callable function
    :rtype: callable
    """
    function_scipy_format = format_function_scipy(function)

    # Create a function def for each starting set in startvals
    if not is_safe(function_scipy_format):
        raise ParsingError('Error while sanitizing input')
    # Sanitizing of function_scipy_format is done so exec use is valid
    # Param_names is sanitized in get_nist_param_names_and_values
    # pylint: disable=exec-used

    local_dict = {}
    global_dict = {'__builtins__': {}, 'np': np}
    exec("def fitting_function(x, " + ','.join(param_names) + "): return "
         + function_scipy_format, global_dict, local_dict)

    return local_dict['fitting_function']


def format_function_scipy(function):
    """
    Formats the function string such that it is scipy-ready.

    :param function: The function to be formatted
    :type function: str

    :return: The formatted function
    :rtype: str
    """

    function = function.replace("exp", "np.exp")
    function = function.replace("^", "**")
    function = function.replace("cos", "np.cos")
    function = function.replace("sin", "np.sin")
    function = function.replace("tan", "np.tan")
    function = function.replace("pi", "np.pi")

    return function


# Due to the nature of this function it is necessary to be able to return at
# multiple places
# pylint: disable=too-many-return-statements, too-many-branches
def is_safe(func_str):
    """
    Verifies that a string is safe to be passed to exec in the context of an
    equation.

    :param func_str: The function to be checked
    :type func_str: string

    :return: Whether the string is of the expected format for an equation
    :rtype: bool
    """
    # Remove whitespace
    func_str = func_str.replace(' ', '')

    # Empty string is safe
    if func_str == '':
        return True

    # These are all safe and can be stripped out
    if 'np' in func_str:
        np_funcs = ['np.exp', 'np.cos', 'np.sin', 'np.tan', 'np.pi']
        for s in np_funcs:
            func_str = func_str.replace(s, '')

    # Store valid symbols for later
    symbols = ['**', '/', '*', '+', '-']

    # Partition on outer brackets
    if '(' in func_str:
        if ')' not in func_str:
            # Number of brackets don't match
            return False

        # Split string "left(centre)right"
        left, remainder = func_str.split('(', 1)
        centre, right = remainder.split(')', 1)
        # Handle nested brackets
        while centre.count('(') != centre.count(')'):
            tmp, right = right.split(')', 1)
            centre = centre + ')' + tmp

        # If left is non-empty it should end with a symbol
        if left != '':
            left_ends_with_symbol = False
            for sym in symbols:
                if left.endswith(sym):
                    left = left.strip(sym)
                    left_ends_with_symbol = True
                    break
            if left_ends_with_symbol is False:
                return False

        # If right is non-empty it should start with a symbol
        if right != '':
            right_starts_with_symbol = False
            for sym in symbols:
                if right.startswith(sym):
                    right = right.strip(sym)
                    right_starts_with_symbol = True
                    break
            if right_starts_with_symbol is False:
                return False

        # Centre should not be empty
        if centre == '':
            return False

        # Return True if all sub parts are safe
        return is_safe(left) and is_safe(centre) and is_safe(right)

    # Split on a symbol and recurse
    for sym in symbols:
        if sym in func_str:
            left, right = func_str.split(sym, 1)

            # Symbol should not be at start or end of string (unless it's a -)
            if (left == '' and sym != '-') or right == '':
                return False

            # Return True if both sub parts are safe
            return is_safe(left) and is_safe(right)

    # Floating points are acceptable
    try:
        float(func_str)
        return True
    except ValueError:
        pass

    # Ints are acceptable
    try:
        int(func_str)
        return True
    except ValueError:
        pass

    # Only remaining acceptable strings are variables
    if func_str[0].isalpha() and func_str.isalnum():
        return True

    # Unparsed output remains
    return False



'''
Test the MINIMIZERS section for the options file
'''
import inspect
import shutil
import os
import unittest

from fitbenchmarking.utils import exceptions
from fitbenchmarking.utils.options import Options


class MininimizerOptionTests(unittest.TestCase):
    """
    Checks that the valid minimizers in the options file are set correctly
    """

    def setUp(self):
        """
        Initializes options class with defaults
        """
        self.options = Options()
        software = ['bumps', 'dfo', 'gsl', 'mantid', 'minuit',
                    'ralfit', 'scipy', 'scipy_ls']
        self.options.software = software

    def test_minimizer_bumps(self):
        """
        Checks valid bumps minimizers are set correctly
        """
        expected = ['amoeba', 'lm-bumps', 'newton', 'mp']
        actual = self.options.minimizers['bumps']
        self.assertEqual(expected, actual)

    def test_minimizer_dfo(self):
        """
        Checks valid dfo minimizers are set correctly
        """
        expected = ['dfogn', 'dfols']
        actual = self.options.minimizers['dfo']
        self.assertEqual(expected, actual)

    def test_minimizer_gsl(self):
        """
        Checks valid gsl minimizers are set correctly
        """
        expected = ['lmsder', 'lmder', 'nmsimplex', 'nmsimplex2',
                    'conjugate_pr', 'conjugate_fr', 'vector_bfgs',
                    'vector_bfgs2', 'steepest_descent']
        actual = self.options.minimizers['gsl']
        self.assertEqual(expected, actual)

    def test_minimizer_mantid(self):
        """
        Checks valid mantid minimizers are set correctly
        """
        expected = ['BFGS',
                    'Conjugate gradient (Fletcher-Reeves imp.)',
                    'Conjugate gradient (Polak-Ribiere imp.)',
                    'Damped GaussNewton', 'Levenberg-Marquardt',
                    'Levenberg-MarquardtMD', 'Simplex',
                    'SteepestDescent', 'Trust Region']
        actual = self.options.minimizers['mantid']
        self.assertEqual(expected, actual)

    def test_minimizer_minuit(self):
        """
        Checks valid minuit minimizers are set correctly
        """
        expected = ['minuit']
        actual = self.options.minimizers['minuit']
        self.assertEqual(expected, actual)

    def test_minimizer_ralfit(self):
        """
        Checks valid ralfit minimizers are set correctly
        """
        expected = ['gn', 'gn_reg', 'hybrid', 'hybrid_reg']
        actual = self.options.minimizers['ralfit']
        self.assertEqual(expected, actual)

    def test_minimizer_scipy(self):
        """
        Checks valid scipy minimizers are set correctly
        """
        expected = ['Nelder-Mead', 'Powell', 'CG', 'BFGS',
                    'Newton-CG', 'L-BFGS-B', 'TNC']
        actual = self.options.minimizers['scipy']
        self.assertEqual(expected, actual)

    def test_minimizer_scipy_ls(self):
        """
        Checks valid scipy_ls minimizers are set correctly
        """
        expected = ['lm-scipy-no-jac', 'lm-scipy', 'trf', 'dogbox']
        actual = self.options.minimizers['scipy_ls']
        self.assertEqual(expected, actual)


class UserMininimizerOptionTests(unittest.TestCase):
    """
    Checks the minimizers in the options file are set correctly or raise errors
    """

    def setUp(self):
        """
        Sets the directory to save the temporary ini files in
        """
        options_dir = os.path.dirname(inspect.getfile(Options))
        self.test_files_dir = os.path.join(options_dir, 'tests', 'files')
        os.mkdir(self.test_files_dir)

    def tearDown(self):
        """
        Deletes temporary folder and results produced
        """
        if os.path.exists(self.test_files_dir):
            shutil.rmtree(self.test_files_dir)

    def generate_user_ini_file(self, options_set, software):
        """
        Generates user defined ini file to be parsed

        :param options_set: option set to be tested
        :type options_set: list
        :param software: software to be tests
        :type software: str

        :return: location of temporary ini file
        :rtype: str
        """
        new_line = '\n{}'.format(" " * (len(software) + 2))
        config_str = \
            "[MINIMIZERS]\n{0}: {1}".format(software,
                                            new_line.join(options_set))
        opts_file = os.path.join(self.test_files_dir,
                                 'test_{}_valid.ini'.format(software))
        with open(opts_file, 'w') as f:
            f.write(config_str)
        return opts_file

    def shared_valid(self, options_set, software):
        """
        Shared test to check that the minimizer option set is valid

        :param options_set: option set to be tested
        :type options_set: list
        :param software: software to be tests
        :type software: str
        """
        opts_file = self.generate_user_ini_file(options_set, software)
        options = Options(opts_file)
        if software not in options.software:
            options.software.append(software)
        actual = options.minimizers[software]
        self.assertEqual(options_set, actual)

    def shared_invalid(self, options_set, software):
        """
        Shared test to check that the minimizer option set is invalid

        :param options_set: option set to be tested
        :type options_set: list
        :param software: software to be tests
        :type software: str
        """
        opts_file = self.generate_user_ini_file(options_set, software)
        with self.assertRaises(exceptions.OptionsError):
            Options(opts_file)

    def test_invalid_option_key(self):
        """
        Tests that the user defined option key is invalid.
        """
        set_option = ['newton', 'de']
        self.shared_invalid(set_option, 'sasview')

    def test_minimizer_bumps_valid(self):
        """
        Checks user set bumps minimizers is valid
        """
        set_option = ['newton', 'de']
        self.shared_valid(set_option, 'bumps')

    def test_minimizer_bumps_invalid(self):
        """
        Checks user set bumps minimizers is invalid
        """
        set_option = ['random_minimizer']
        self.shared_invalid(set_option, 'bumps')

    def test_minimizer_dfo_valid(self):
        """
        Checks user set dfo minimizers is valid
        """
        set_option = ['dfogn', 'dfols']
        self.shared_valid(set_option, 'dfo')

    def test_minimizer_dfo_invalid(self):
        """
        Checks user set dfo minimizers is invalid
        """
        set_option = ['CG']
        self.shared_invalid(set_option, 'dfo')

    def test_minimizer_gsl_valid(self):
        """
        Checks user set gsl minimizers is valid
        """
        set_option = ['lmsder', 'lmder', 'nmsimplex']
        self.shared_valid(set_option, 'gsl')

    def test_minimizer_gsl_invalid(self):
        """
        Checks user set gsl minimizers is invalid
        """
        set_option = ['newton']
        self.shared_invalid(set_option, 'gsl')

    def test_minimizer_mantid_valid(self):
        """
        Checks user set mantid minimizers is valid
        """
        set_option = ['Damped GaussNewton', 'Levenberg-Marquardt',
                      'Levenberg-MarquardtMD']
        self.shared_valid(set_option, 'mantid')

    def test_minimizer_mantid_invalid(self):
        """
        Checks user set mantid minimizers is invalid
        """
        set_option = ['lm-mantid', 'Levenberg-Marquardt',
                      'Levenberg-MarquardtMD']
        self.shared_invalid(set_option, 'mantid')

    def test_minimizer_minuit_valid(self):
        """
        Checks user set minuit minimizers is valid
        """
        set_option = ['minuit']
        self.shared_valid(set_option, 'minuit')

    def test_minimizer_minuit_invalid(self):
        """
        Checks user set minuit minimizers is invalid
        """
        set_option = ['line search']
        self.shared_invalid(set_option, 'minuit')

    def test_minimizer_ralfit_valid(self):
        """
        Checks user set ralfit minimizers is valid
        """
        set_option = ['hybrid_reg']
        self.shared_valid(set_option, 'ralfit')

    def test_minimizer_ralfit_invalid(self):
        """
        Checks user set ralfit minimizers is invalid
        """
        set_option = ['Trust_region']
        self.shared_invalid(set_option, 'ralfit')

    def test_minimizer_scipy_valid(self):
        """
        Checks user set scipy minimizers is valid
        """
        set_option = ['Nelder-Mead', 'TNC', 'BFGS', 'CG']
        self.shared_valid(set_option, 'scipy')

    def test_minimizer_scipy_invalid(self):
        """
        Checks user set scipy minimizers is invalid
        """
        set_option = ['Nelder-Mead', 'TNC-scipy', 'BFGS', 'CG-NE']
        self.shared_invalid(set_option, 'scipy')

    def test_minimizer_scipy_ls_valid(self):
        """
        Checks user set scipy_ls minimizers is valid
        """
        set_option = ['lm-scipy', 'trf']
        self.shared_valid(set_option, 'scipy_ls')

    def test_minimizer_scipy_ls_invalid(self):
        """
        Checks user set scipy_ls minimizers is invalid
        """
        set_option = ['lm-minpack-jac', 'trf']
        self.shared_invalid(set_option, 'scipy_ls')



"""
Tests for the log file.

Note: As get_logger is effectively an alias, it is not tested here.
"""
try:
    from tempfile import TemporaryDirectory
except ImportError:
    from backports.tempfile import TemporaryDirectory
import os
from unittest import TestCase

from fitbenchmarking.utils.log import get_logger, setup_logger
from fitbenchmarking.utils.output_grabber import OutputGrabber


class TestSetupLogger(TestCase):
    """
    Tests for the setup_logger method in log.
    """

    def setUp(self):
        self.temp_dir = TemporaryDirectory()
        self.log_file = os.path.join(self.temp_dir.name, 'some_log.txt')

    def test_file_creation_correct(self):
        """
        Test that the logger creates the correct file.
        """
        self.assertFalse(os.path.exists(self.log_file),
                         'Log file already present?')
        setup_logger(log_file=self.log_file, name='test1')
        logger = get_logger('test1')
        logger.info('Works?')
        self.assertTrue(os.path.exists(self.log_file),
                        'Failed to create log file.')

    def test_append_mode_on(self):
        """
        Test that append mode works when set to true.
        """
        setup_logger(log_file=self.log_file, name='test2')
        logger = get_logger('test2')
        logger.info('Before reset')
        setup_logger(log_file=self.log_file, append=True, name='test2')
        logger.info('After reset')

        with open(self.log_file, 'r') as f:
            lines = f.readlines()

        self.assertEqual(len(lines), 2, 'Wrong number of lines in log file')
        self.assertIn('Before reset', lines[0])
        self.assertIn('After reset', lines[1])

    def test_append_mode_off(self):
        """
        Test that append mode works when set to false.
        """
        setup_logger(log_file=self.log_file, name='test3')

        logger = get_logger('test3')
        logger.info('Before reset')
        setup_logger(log_file=self.log_file, append=False, name='test3')
        logger.info('After reset')

        with open(self.log_file, 'r') as f:
            lines = f.readlines()

        self.assertEqual(len(lines), 1, 'Wrong number of lines in log file')
        self.assertNotIn('Before reset', lines[0])
        self.assertIn('After reset', lines[0])


def test_level(capsys):
    """
    Test that setting the level limits the output from the logger.
    Note: This relies on a pytest fixture (capsys)
    """
    temp_dir = TemporaryDirectory()
    log_file = os.path.join(temp_dir.name, 'some_log.txt')
    setup_logger(log_file=log_file, level='WARNING', append=True, name='test4')
    logger = get_logger('test4')

    logger.error('An error message')
    logger.warning('A warning message')
    logger.info('An info message')

    out, _ = capsys.readouterr()
    output = out.splitlines()

    assert (len(output) == 2), 'Wrong number of lines in log'
    assert ('An error message' in output[0]), 'Error not captured by log'
    assert ('A warning message' in output[1]), 'Warning not captured by log'



"""
Implements the FittingProblem class, this will be the object that inputs are
parsed into before being passed to the controllers
"""

from __future__ import (absolute_import, division, print_function)

try:
    from itertools import izip_longest
except ImportError:
    # python3
    from itertools import zip_longest as izip_longest
import numpy as np

from fitbenchmarking.utils.exceptions import FittingProblemError


# Using property getters and setters means that the setter does not always use
# self
# pylint: disable=no-self-use
class FittingProblem:
    r"""
    Definition of a fitting problem, which will be populated by a parser from a
    problem definition file.

    This defines a fitting problem where, given a set of :math:`n` data points
    :math:`(x_i,y_i)`, associated errors :math:`e_i`, and a model
    function :math:`f(x,p)`, we find the optimal parameters in the
    least-squares sense by solving:

    .. math:: \min_p \sum_{i=1}^n \left( \frac{y_i - f(x_i, p)}{e_i} \right)^2

    where :math:`p` is a vector of length :math:`m`, and we start from a given
    intial guess for the optimal parameters.
    """

    def __init__(self, options):
        """
        Initialises variable used for temporary storage.

        :param options: all the information specified by the user
        :type options: fitbenchmarking.utils.options.Options
        """
        self.options = options
        #: *string* Name (title) of the fitting problem
        self.name = None

        #: *string* Equation (function or model) to fit against data
        self.equation = None

        #: *float* The start of the range to fit model data over
        #: (if different from entire range)
        self.start_x = None

        #: *float* The end of the range to fit model data over
        #: (if different from entire range) (/float/)
        self.end_x = None

        #: *numpy array* The x-data
        self.data_x = None

        #: *numpy array* The y-data
        self.data_y = None

        #: *numpy array* The errors
        self.data_e = None

        #: *list of dict*
        #: Starting values of the fitting parameters
        #:
        #: e.g.
        #: :code:`[{p1_name: p1_val1, p2_name: p2_val1, ...},
        #: {p1_name: p1_val2, ...}, ...]`
        self.starting_values = None

        #: *dict*
        #: Smallest and largest values of interest in the data
        #:
        #: e.g.
        #: :code:`{p1_name: [p1_min, p1_max], ...}`
        self.value_ranges = None

        #: Callable function
        self.function = None

        self._param_names = None

        #: *numpy array* The index for sorting the data (used in plotting)
        self.sorted_index = None

        #: *dict*
        #: Container for software specific information.
        #: This should be avoided if possible.
        self.additional_info = {}

        #: *bool* Used to check if a problem is using multifit.
        self.multifit = False

        # Used to assign the Jacobian class in
        # fitbenchmarking/core/fitting_benchmarking.py
        self.jac = None

    @property
    def param_names(self):
        """
        Utility function to get the parameter names

        :return: the names of the parameters
        :rtype: list of str
        """
        if self._param_names is None:
            # pylint: disable=unsubscriptable-object
            self._param_names = list(self.starting_values[0].keys())
        return self._param_names

    @param_names.setter
    def param_names(self, value):
        raise FittingProblemError('param_names should not be edited')

    def eval_f(self, params, x=None):
        """
        Function evaluation method

        :param params: parameter value(s)
        :type params: list
        :param x: x data values or None, if None this uses self.data_x
        :type x: numpy array

        :return: y data values evaluated from the function of the problem
        :rtype: numpy array
        """
        if self.function is None:
            raise FittingProblemError('Cannot call function before setting '
                                      'function.')
        if x is None:
            x = self.data_x
        # pylint: disable=not-callable
        out = self.function(x, *params)
        return out

    def eval_r(self, params, x=None, y=None, e=None):
        """
        Calculate residuals and weight them if using errors

        :param params: The parameters to calculate residuals for
        :type params: list
        :param x: x data points, defaults to self.data_x
        :type x: numpy array, optional
        :param y: y data points, defaults to self.data_y
        :type y: numpy array, optional
        :param e: error at each data point, defaults to self.data_e
        :type e: numpy array, optional

        :return: The residuals for the datapoints at the given parameters
        :rtype: numpy array
        """

        if x is None and y is None and e is None:
            x = self.data_x
            y = self.data_y
            e = self.data_e
        elif x is None or y is None:
            raise FittingProblemError('Residuals could not be computed with '
                                      'only one of x and y.')

        result = y - self.eval_f(params=params, x=x)
        if e is not None:
            result = result / e
        return result

    def eval_r_norm(self, params, x=None, y=None, e=None):
        """
        Evaluate the square of the L2 norm of the residuals

        :param params: The parameters to calculate residuals for
        :type params: list
        :param x: x data points, defaults to self.data_x
        :type x: numpy array, optional
        :param y: y data points, defaults to self.data_y
        :type y: numpy array, optional
        :param e: error at each data point, defaults to self.data_e
        :type e: numpy array, optional

        :return: The sum of squares of residuals for the datapoints at the
                 given parameters
        :rtype: numpy array
        """
        r = self.eval_r(params=params, x=x, y=y, e=e)
        return np.dot(r, r)

    def eval_starting_params(self, param_set):
        """
        Evaluate the function using the starting parameters.

        :param param_set: The index of the parameter set in starting_params
        :type param_set: int

        :return: Results from evaluation
        :rtype: numpy array
        """
        if self.starting_values is None:
            raise FittingProblemError('Cannot call function before setting '
                                      'starting values.')
        # pylint: disable=unsubscriptable-object
        return self.eval_f(self.starting_values[param_set].values())

    def get_function_params(self, params):
        """
        Return the function definition in a string format for output

        :param params: The parameters to use in the function string
        :type params: list

        :return: Representation of the function
                 example format: 'b1 * (b2+x) | b1=-2.0, b2=50.0'
        :rtype: string
        """
        params = ['{}={}'.format(n, p) for n, p
                  in izip_longest(self.param_names,
                                  params if params is not None else [])]
        param_string = ', '.join(params)

        return param_string

    def verify(self):
        """
        Basic check that minimal set of attributes have been set.

        Raise FittingProblemError if object is not properly initialised.
        """
        values = {'data_x': np.ndarray,
                  'data_y': np.ndarray,
                  'starting_values': list}

        for attr_name, attr_type in values.items():
            attr = getattr(self, attr_name)
            type_match = isinstance(attr, attr_type)
            try:
                type_match = type_match or isinstance(attr[0], attr_type)  # NOQA; pylint: disable=unsubscriptable-object
            except (TypeError, IndexError):
                pass
            if not type_match:
                raise FittingProblemError(
                    'Attribute "{}" is not the expected type. Expected "{}", '
                    'got "{}".'.format(attr_name, attr_type, type(attr)))
        if self.function is None:
            raise FittingProblemError('Attribute "function" has not been set.')

    def correct_data(self):
        """
        Strip data that overruns the start and end x_range,
        and approximate errors if not given.
        Modifications happen on member variables.
        """
        if not self.multifit:
            correct_vals = correct_data(x=self.data_x,
                                        y=self.data_y,
                                        e=self.data_e,
                                        startx=self.start_x,
                                        endx=self.end_x,
                                        use_errors=self.options.use_errors)
            self.data_x = correct_vals[0]
            self.data_y = correct_vals[1]
            self.data_e = correct_vals[2]
            self.sorted_index = correct_vals[3]
        else:
            # Mantid multifit problem
            self.sorted_index = []
            for i in range(len(self.data_x)):
                correct_vals = correct_data(x=self.data_x[i],
                                            y=self.data_y[i],
                                            e=self.data_e[i],
                                            startx=self.start_x[i],
                                            endx=self.end_x[i],
                                            use_errors=self.options.use_errors)
                self.data_x[i] = correct_vals[0]
                self.data_y[i] = correct_vals[1]
                self.data_e[i] = correct_vals[2]
                self.sorted_index.append(correct_vals[3])


def correct_data(x, y, e, startx, endx, use_errors):
    """
    Strip data that overruns the start and end x_range,
    and approximate errors if not given.

    :param x: x data
    :type x: np.array
    :param y: y data
    :type y: np.array
    :param e: error data
    :type e: np.array
    :param startx: minimum x value
    :type startx: float
    :param endx: maximum x value
    :type endx: float
    :param use_errors: whether errors should be added if not present
    :type use_errors: bool
    """
    # fix data_e
    if use_errors:
        if e is None:
            e = np.sqrt(abs(y))

        # The values of data_e are used to divide the residuals.
        # If these are (close to zero), then this blows up.
        # This is particularly a problem if trying to fit
        # counts, which may follow a Poisson distribution.
        #
        # Fix this by cutting values less than a certain value
        trim_value = 1.0e-8
        e[e < trim_value] = trim_value
    else:
        e = None

    # impose x ranges
    # pylint: disable=no-member, assignment-from-no-return
    if startx is not None and endx is not None:
        mask = np.logical_and(x >= startx,
                              x <= endx)
        x = x[mask]
        y = y[mask]
        if e is not None:
            e = e[mask]

    # Stores the indices of the sorted data
    sorted_index = np.argsort(x)

    return x, y, e, sorted_index



"""
This file holds all FitBenchmarking exceptions, organised by exception id
"""


class FitBenchmarkException(Exception):
    """
    The base class for all FitBenchmarking exceptions

    To define a new exception, inherrit from this and override the
    _class_message
    """

    def __init__(self, message=''):
        super(FitBenchmarkException, self).__init__(message)
        self._class_message = 'An unknown exception occurred.'
        self._obj_message = message
        self.error_code = 1
        self._error_string = None

    def __str__(self):
        if self._obj_message != '':
            self._error_string = '{}\nDetails: {}'.format(self._class_message,
                                                          self._obj_message)
        else:
            self._error_string = self._class_message

        return self._error_string.strip()


class OptionsError(FitBenchmarkException):
    """
    Indicates an error during processing options.
    """

    def __init__(self, message=''):
        super(OptionsError, self).__init__(message)

        self._class_message = 'Failed to process options.'
        self.error_code = 2


class ParsingError(FitBenchmarkException):
    """
    Indicates an error during parsing.
    """

    def __init__(self, message=''):
        super(ParsingError, self).__init__(message)

        self._class_message = 'Could not parse problem.'
        self.error_code = 3


class NoParserError(FitBenchmarkException):
    """
    Indicates a parser could not be found.
    """

    def __init__(self, message=''):
        super(NoParserError, self).__init__(message)

        self._class_message = 'Could not find parser.'
        self.error_code = 4


class MissingSoftwareError(FitBenchmarkException):
    """
    Indicates that the requirements for a software package are not available.
    """

    def __init__(self, message=''):
        super(MissingSoftwareError, self).__init__(message)

        self._class_message = 'Missing dependencies for fit.'
        self.error_code = 5


class NoControllerError(FitBenchmarkException):
    """
    Indicates a controller could not be found
    """

    def __init__(self, message=''):
        super(NoControllerError, self).__init__(message)

        self._class_message = 'Could not find controller.'
        self.error_code = 6


class ControllerAttributeError(FitBenchmarkException):
    """
    Indicates an issue with the attributes within a controller
    """

    def __init__(self, message=''):
        super(ControllerAttributeError, self).__init__(message)

        self._class_message = 'Error in the controller attributes.'
        self.error_code = 7


class NoDataError(FitBenchmarkException):
    """
    Indicates that no data could be found.
    """

    def __init__(self, message=''):
        super(NoDataError, self).__init__(message)

        self._class_message = 'No data found.'
        self.error_code = 8


class UnknownMinimizerError(FitBenchmarkException):
    """
    Indicates that the controller does not support a given minimizer given
    the current "algorithm_type" option set.
    """

    def __init__(self, message=''):
        super(UnknownMinimizerError, self).__init__(message)

        self._class_message = 'Minimizer cannot be run with Controller with ' \
                              'current "algorithm_type" option set.'
        self.error_code = 9


class FittingProblemError(FitBenchmarkException):
    """
    Indicates a problem with the fitting problem.
    """

    def __init__(self, message=''):
        super(FittingProblemError, self).__init__(message)

        self._class_message = 'Fitting Problem raised and exception.'
        self.error_code = 10


class NoJacobianError(FitBenchmarkException):
    """
    Indicates a problem with the Jacobian import.
    """

    def __init__(self, message=''):
        super(NoJacobianError, self).__init__(message)

        self._class_message = 'Could not find Jacobian class'
        self.error_code = 11


class UnknownTableError(FitBenchmarkException):
    """
    Indicates a problem with the fitting problem.
    """

    def __init__(self, message=''):
        super(UnknownTableError, self).__init__(message)

        self._class_message = 'Set table option could not be found'
        self.error_code = 12


class NoResultsError(FitBenchmarkException):
    """
    Indicates a problem with the fitting problem.
    """

    def __init__(self, message=''):
        super(NoResultsError, self).__init__(message)

        self._class_message = 'FitBenchmarking ran with no results'
        self.error_code = 13


class UnsupportedMinimizerError(FitBenchmarkException):
    """
    Indicates that the controller does not support a given minimizer.
    """

    def __init__(self, message=''):
        super(UnsupportedMinimizerError, self).__init__(message)

        self._class_message = 'FitBenchmarking ran with no results'
        self.error_code = 14



from __future__ import (absolute_import, division, print_function)

import inspect
import os
try:
    from tempfile import TemporaryDirectory
except ImportError:
    from backports.tempfile import TemporaryDirectory
import unittest

import fitbenchmarking
from fitbenchmarking.parsing.fitting_problem import FittingProblem
from fitbenchmarking.jacobian.SciPyFD_2point_jacobian import ScipyTwoPoint
from fitbenchmarking.results_processing import support_page
from fitbenchmarking.utils.fitbm_result import FittingResult
from fitbenchmarking.utils.options import Options


class CreateTests(unittest.TestCase):

    def setUp(self):
        self.options = Options()
        problems = []
        for i in range(5):
            problem = FittingProblem(self.options)
            problem.name = 'prob {}'.format(i)
            problem.starting_values = [{'x': 1}]
            problems.append(problem)

        minimizers = ['min_a', 'min_b', 'min_c']
        self.results = [[FittingResult(options=self.options,
                                       problem=p,
                                       jac=ScipyTwoPoint(p),
                                       initial_params=[],
                                       params=[],
                                       minimizer=m)
                         for m in minimizers]
                        for p in problems]
        root = os.path.dirname(inspect.getfile(fitbenchmarking))
        self.dir = TemporaryDirectory(dir=root)

    def test_create_unique_files(self):
        """
        Tests that the create function creates a set of unique files.
        """
        support_page.create(results_per_test=self.results,
                            group_name='test_group',
                            support_pages_dir=self.dir.name,
                            options=self.options)

        file_names = sorted([r.support_page_link
                             for pr in self.results
                             for r in pr])

        unique_names = sorted(list(set(file_names)))

        self.assertListEqual(unique_names, file_names)


class CreateProbGroupTests(unittest.TestCase):
    """
    Tests that the correct files are created by group tests.
    Does not test the content of the file currently.
    """

    def setUp(self):
        self.options = Options()
        problem = FittingProblem(self.options)
        problem.name = 'prob a'
        problem.equation = 'equation!'
        problem.starting_values = [{'x': 1}]

        minimizers = ['min_a', 'min_b', 'min_c']
        self.results = [FittingResult(options=self.options,
                                      problem=problem,
                                      jac=ScipyTwoPoint(problem),
                                      initial_params=[],
                                      params=[],
                                      minimizer=m)
                        for m in minimizers]

        root = os.path.dirname(inspect.getfile(fitbenchmarking))
        self.dir = TemporaryDirectory(dir=root)

    def test_create_files(self):
        """
        Tests that files are created for each result.
        """
        support_page.create_prob_group(prob_results=self.results,
                                       group_name='test_group',
                                       support_pages_dir=self.dir.name,
                                       options=self.options)
        self.assertTrue(all(
            os.path.exists(r.support_page_link) for r in self.results))

    def test_file_name(self):
        """
        Tests that the filenames are in the expected form.
        """
        support_page.create_prob_group(prob_results=self.results,
                                       group_name='test_group',
                                       support_pages_dir=self.dir.name,
                                       options=self.options)
        file_names = [r.support_page_link for r in self.results]
        expected = [os.path.join(os.path.relpath(self.dir.name), f)
                    for f in ['test_group_prob_a_min_a.html',
                              'test_group_prob_a_min_b.html',
                              'test_group_prob_a_min_c.html']]

        self.assertListEqual(file_names, expected)


class GetFigurePathsTests(unittest.TestCase):
    """
    Tests the very simple get_figure_paths function
    """

    def setUp(self):
        self.options = Options()
        problem = FittingProblem(self.options)
        problem.name = 'prob a'
        problem.equation = 'equation!'
        problem.starting_values = [{'x': 1}]
        self.result = FittingResult(options=self.options,
                                    problem=problem,
                                    jac=ScipyTwoPoint(problem),
                                    initial_params=[],
                                    params=[],
                                    minimizer='test')

    def test_with_links(self):
        """
        Tests that the returned links are correct when links are passed in.
        """
        self.result.figure_link = 'some_link'
        self.result.start_figure_link = 'other_link'
        figure_link, start_link = support_page.get_figure_paths(self.result)
        self.assertEqual(figure_link, os.path.join('figures', 'some_link'))
        self.assertEqual(start_link, os.path.join('figures', 'other_link'))

    def test_no_links(self):
        """
        Tests that links are not changed if an empty string is given.
        """
        self.result.figure_link = ''
        self.result.start_figure_link = ''
        figure_link, start_link = support_page.get_figure_paths(self.result)
        self.assertEqual(figure_link, '')
        self.assertEqual(start_link, '')


if __name__ == "__main__":
    unittest.main()



"""
Implements a controller for the CERN pacakage Minuit
https://seal.web.cern.ch/seal/snapshot/work-packages/mathlibs/minuit/
using the iminuit python interface
http://iminuit.readthedocs.org
"""
from iminuit import Minuit
import numpy as np

from fitbenchmarking.controllers.base_controller import Controller


class MinuitController(Controller):
    """
    Controller for the Minuit fitting software
    """

    def __init__(self, problem):
        """
        Initializes variable used for temporary storage.

        :param problem: Problem to fit
        :type problem: FittingProblem
        """
        super(MinuitController, self).__init__(problem)
        self._popt = None
        self._initial_step = None
        self._minuit_problem = None
        self.algorithm_check = {
            'all': ['minuit'],
            'ls': [None],
            'deriv_free': [None],
            'general': ['minuit']}

    def setup(self):
        """
        Setup for Minuit
        """
        # minuit requires an initial step size.
        # The docs say
        # "A good guess is a fraction of the initial
        #  fit parameter value, e.g. 10%
        #  (be careful when applying this rule-of-thumb
        #  when the initial parameter value is zero "
        self._initial_step = 0.1 * np.array(self.initial_params)
        # set small steps to something sensible(?)
        self._initial_step[self._initial_step < 1e-12] = 1e-12
        self._minuit_problem = Minuit.from_array_func(self.problem.eval_r_norm,
                                                      self.initial_params,
                                                      error=self._initial_step,
                                                      errordef=1)

    def fit(self):
        """
        Run problem with Minuit
        """
        self._minuit_problem.migrad()  # run optimizer
        self._status = 0 if self._minuit_problem.migrad_ok() else 1

    def cleanup(self):
        """
        Convert the result to a numpy array and populate the variables results
        will be read from
        """
        fmin = self._minuit_problem.get_fmin()
        if self._status == 0:
            self.flag = 0
        elif fmin.has_reached_call_limit:
            self.flag = 1
        else:
            self.flag = 2

        self._popt = self._minuit_problem.np_values()
        self.final_params = self._popt



"""
compare table
"""
import numpy as np
import os
from fitbenchmarking.results_processing.base_table import Table


class CompareTable(Table):
    """

    The combined results show the accuracy in the first line of the cell and
    the runtime on the second line of the cell.

    """

    def __init__(self, results, best_results, options, group_dir,
                 pp_locations, table_name):
        """
        Initialise the compare table which shows both accuracy and runtime
        results

        :param results: results nested array of objects
        :type results: list of list of
                       fitbenchmarking.utils.fitbm_result.FittingResult
        :param best_results: best result for each problem
        :type best_results: list of
                        fitbenchmarking.utils.fitbm_result.FittingResult
        :param options: Options used in fitting
        :type options: utils.options.Options
        :param group_dir: path to the directory where group results should be
                          stored
        :type group_dir: str
        :param pp_locations: tuple containing the locations of the
                             performance profiles (acc then runtime)
        :type pp_locations: tuple(str,str)
        :param table_name: Name of the table
        :type table_name: str
        """
        self.name = 'compare'
        super(CompareTable, self).__init__(results, best_results, options,
                                           group_dir, pp_locations, table_name)

        self.has_pp = True
        self.pp_filenames = \
                        [os.path.relpath(pp,group_dir) for pp in pp_locations]

    def get_values(self, results_dict):
        """
        Gets the main values to be reported in the tables

        :param results_dict: dictionary containing results where the keys
                             are the problem sets and the values are lists
                             of results objects
        :type results_dict: dictionary

        :return: First dictionary containing the absolute chi_sq and runtime
                 results and the second contains the normalised chi_sq and
                 runtime results with respect to the smallest values.
        :rtype: tuple(dict, dict)
        """
        abs_value = {}
        rel_value = {}
        for key, value in results_dict.items():
            acc_abs_value = [v.chi_sq for v in value]
            acc_min_value = np.min(acc_abs_value)
            acc_rel_value = [v.chi_sq / acc_min_value for v in value]

            runtime_abs_value = [v.runtime for v in value]
            runtime_min_value = np.min(runtime_abs_value)
            runtime_rel_value = [v.runtime / runtime_min_value for v in value]

            abs_value[key] = [acc_abs_value, runtime_abs_value]
            rel_value[key] = [acc_rel_value, runtime_rel_value]

        return abs_value, rel_value

    def display_str(self, results):
        """
        Function that combines the accuracy and runtime results into the
        string representation.

        :param results: tuple containing absolute and relative values
        :type results: tuple

        :return: dictionary containing the string representation of the values
                 in the table.
        :rtype: dict
        """
        abs_results, rel_results = results
        comp_mode = self.options.comparison_mode
        result_template = self.output_string_type[self.options.comparison_mode]
        table_output = {}
        for key in abs_results.keys():
            acc_abs, runtime_abs = abs_results[key]
            acc_rel, runtime_rel = rel_results[key]
            if comp_mode == "abs":
                table_output[key] = \
                    [result_template.format(acc) + '<br>'
                     + result_template.format(runtime)
                     for acc, runtime in zip(acc_abs, runtime_abs)]
            elif comp_mode == "rel":
                table_output[key] = \
                    [result_template.format(acc) + '<br>'
                     + result_template.format(runtime)
                     for acc, runtime in zip(acc_rel, runtime_rel)]
            elif comp_mode == "both":
                table_output[key] = \
                    [result_template.format(acc_a, acc_r) + '<br>'
                     + result_template.format(runtime_a, runtime_r)
                     for acc_a, acc_r, runtime_a, runtime_r in
                     zip(acc_abs, acc_rel, runtime_abs, runtime_rel)]
        return table_output

    def colour_highlight(self, value, colour):
        """
        Enable HTML colours in the HTML output.

        :param value: Row data from the pandas array
        :type value: pandas.core.series.Series
        :param colour: dictionary containing error codes from the minimizers
        :type colour: dict

        :return: list of HTML colours
        :rtype: list
        """
        color_template = 'background-image: linear-gradient({0},{0},{1},{1})'
        name = value.name.split('"')[2].replace("</a>", "")[1:]
        output_colour = []
        acc_colour, runtime_colour = colour[name]
        for acc, runtime in zip(acc_colour, runtime_colour):
            output_colour.append(color_template.format(acc, runtime))
        return output_colour



"""
This file calls the pycutest interface for SIF data
"""

from __future__ import print_function

import os
from collections import OrderedDict

import numpy as np
import pycutest

from fitbenchmarking.parsing.base_parser import Parser
from fitbenchmarking.parsing.fitting_problem import FittingProblem
from fitbenchmarking.utils.exceptions import ParsingError

try:
    from tempfile import TemporaryDirectory
except ImportError:
    from backports.tempfile import TemporaryDirectory

if os.path.isdir(os.environ["PYCUTEST_CACHE"]+"/pycutest_cache_holder"):
    # clear the cache
    for cached_problem in pycutest.all_cached_problems():
        pycutest.clear_cache(cached_problem[0], cached_problem[1])


# pylint: disable=attribute-defined-outside-init, too-few-public-methods

class CutestParser(Parser):
    """
    Use the pycutest interface to parse SIF files

    This parser has some quirks.
    Due to the complex nature of SIF files, we utilise pycutest to generate the
    function. This function returns the residual `r(f,x,y,e) := (f(x) - y)/e`
    for some parameters x.
    For consistency we require the value to be `f(x, p)`.

    To avoid having to convert back from the residual in the evaluation, we
    store the data separately and set y to 0.0, and e to 1.0 for all
    datapoints.

    We then accomodate for the missing x argument by caching x against an
    associated function `f_x(p)`.

    As such the function is defined by: `f(x, p) := r(f_x,p,0,1)`

    If a new x is passed, we write a new file and parse it to generate a
    function. We define x to be new if ``not np.isclose(x, cached_x)`` for each
    ``cached_x`` that has already been stored.
    """

    def parse(self):
        """
        Get data into a Fitting Problem via cutest.

        :return: The fully parsed fitting problem
        :rtype: fitbenchmarking.parsing.fitting_problem.FittingProblem
        """
        self.mastsif_dir = TemporaryDirectory()

        # set the MASTSIF environment variable so that pycutest
        # can find the sif files
        os.environ["MASTSIF"] = self.mastsif_dir.name

        self._num_params = None

        # get just the short filename (minus the .SIF)
        fp = FittingProblem(self.options)

        # Collect x and create new file with blank y
        fname, fp.data_x, fp.data_y, fp.data_e = self._setup_data()

        self._p = _import_problem(fname)

        fp.name = self._p.name

        fp.function = self._function  # self._p.objcons
        fp.equation = None
        fp.starting_values = self._get_starting_values()
        fp.start_x = None
        fp.end_x = None

        # Create a list of x and f.
        # If a new x is given we will create and parse a new file
        self._cache = [(fp.data_x, self._p.objcons)]

        return fp

    def _function(self, x, *params):
        """
        If these x values have been passed in before, then run the function
        Otherwise, create a new problem file, parse and cache the function

        :param x: The data to evaluate at
        :type x: np.array
        :return: The result of evaluating at the given x
        :rtype: np.array
        """
        os.environ["MASTSIF"] = self.mastsif_dir.name

        for cx, cf in self._cache:
            if np.isclose(cx, x).all():
                f = cf
                break
        else:
            fname, _, _, _ = self._setup_data(x)
            p = _import_problem(fname)
            f = p.objcons
            self._cache.append((x, f))
        _, fx = f(np.asarray(params))

        return fx

    def _get_starting_values(self):

        starting_values = [
            OrderedDict([
                ('f{}'.format(i), self._p.x0[i])
                for i in range(self._num_params)
            ])
        ]

        return starting_values

    def _setup_data(self, x=None):
        """
        Reads datapoints from CUTEst SIF files and rewrites values where
        needed.

        With x=None, read all data from file into x_data, y_data, and e_data,
        then create a new file with y values of 0.0, and e values of 1.0.

        If x is given, create a file with the given x data, 0.0 y values, and
        1.0 e values, and return the path to the file.

        y and e are set here to avoid the function returning residuals that we
        then have to convert every time the function is evaluated.

        :param x: X values to save (used for evaluating at new points)
        :type x: np.array
        :returns: path to new sif file, data_x, data_y, data_e
        :rtype: str, numpy.ndarray, numpy.ndarray, numpy.ndarray
        """

        if self.file.closed:
            with open(self._filename, 'r') as f:
                lines = f.readlines()
        else:
            lines = self.file.readlines()

        if x is None:
            x, y, e, to_write, n = _read_x(lines)
            self._num_params = n
        else:
            if not x.shape:
                x = np.array([x])
            x, y, e, to_write = _write_x(lines, x)

        file_path = os.path.join(self.mastsif_dir.name,
                                 '{}.SIF'.format(str(id(x))[-10:]))

        with open(file_path, 'w') as f:
            f.writelines(to_write)

        return file_path, x, y, e


def _read_x(lines):
    """
    Read data from the list of lines from the file.
    Overwrite the Y and E values with 0.0 and 1.0 respectively, and return a
    new list of lines of text.

    :param lines: The text to parse data from.
    :type lines: list of str
    :return: x, y, and error data, list of text to write, and number of
             parameters from the file.
    :rtype: numpy array, numpy array, numpy array, list of str, int
    """
    to_write = []
    # SIF requires columns of 25 chars, so line[:col_width-1] will be 1 column
    col_width = 25

    x_idx, y_idx, e_idx = 0, 0, 0

    for line in lines:
        if "IE M " in line:
            data_count = int(line.split()[2])
            # this will always come before x/y data so allocate space now
            data_x = np.zeros(data_count)
            data_y = np.zeros(data_count)
            data_e = np.zeros(data_count)
        elif "IE N " in line:
            num_params = int(line.split()[2])
        elif "RE X" in line:
            data_x[x_idx] = float(line.split()[2])
            x_idx += 1
        elif "RE Y" in line:
            data_y[y_idx] = float(line.split()[2])
            y_idx += 1
            line = line[:col_width-1] + '0.0'
        elif 'RE E' in line:
            data_e[e_idx] = float(line.split()[2])
            e_idx += 1
            line = line[:col_width-1] + '1.0'
        to_write.append(line + '\n')

    _check_data(data_count, x_idx, y_idx, e_idx)
    if not e_idx:
        data_e = None
    return data_x, data_y, data_e, to_write, num_params


def _write_x(lines, x):
    """
    Overwrite x values in the list of lines from the file with values from a
    numpy array.
    Use Y=0.0 and E=1.0, and return the new list of lines of text.

    :param lines: The text to parse data from.
    :type lines: list of str
    :param x: The data to write in place of the data in file
    :type x: numpy array
    :return: x, y, and error data, list of text to write.
    :rtype: numpy array, numpy array, numpy array, list of str
    """
    to_write = []
    # SIF requires columns of 25 chars, so line[:col_width-1] will be 1 column
    col_width = 25

    written = False

    for line in lines:
        if "IE M " in line:
            line = line[:col_width-1] + str(len(x))
        elif "IE MLOWER" in line:
            line = line[:col_width-1] + '1\n'
        elif "IE MUPPER" in line:
            line = line[:col_width-1] + str(len(x))
        elif "RE X" in line:
            # Only write data once.
            # Once written can skip all x entries
            if written:
                continue
            line = ''
            new_lines = []
            for i, val in enumerate(x):
                idx = i+1
                spacing = ' '*(col_width - (5 + len(str(idx))))
                new_lines.extend([' RE X{}{}{}'.format(idx, spacing, val),
                                  ' RE Y{}{}0.0'.format(idx, spacing),
                                  ' RE E{}{}1.0'.format(idx, spacing)])
            x_idx = y_idx = e_idx = len(new_lines) / 3
            line = '\n'.join(new_lines)
            written = True
        elif "RE Y" in line or 'RE E' in line:
            # These should be ignored as they are written when X is found.
            continue
        to_write.append(line + '\n')

    _check_data(len(x), x_idx, y_idx, e_idx)
    return x, np.zeros_like(x), np.ones_like(x), to_write


def _check_data(count, x, y, e):
    """
    Check that then number of data points for x, y, and errors are consistent.

    :param count: Expected number of datapoints
    :type count: int
    :param x: Number of x data points
    :type x: int
    :param y: Number of y data points
    :type y: int
    :param e: Number of error data points
    :type e: int
    :raises ParsingError: If x, y, or e does not match count. (e can also be 0)
    """

    if x != count:
        raise ParsingError('Wrong number of x data points. Got {}, '
                           'expected {}'.format(x, count))
    if y != count:
        raise ParsingError('Wrong number of y data points. Got {}, '
                           'expected {}'.format(y, count))
    if e not in (0, count):
        raise ParsingError('Wrong number of e data points. Got {}, '
                           'expected {}'.format(e, count))


def _import_problem(file_name):
    """
    Import the problem using cutest

    :param file_name: The path to the file
    :type file_name: str
    :return: The parsed problem
    :rtype: pycutest.CUTEstProblem
    """

    problem_ext = os.path.basename(file_name)
    problem, _ = os.path.splitext(problem_ext)

    return pycutest.import_problem(problem)



# Tests for the controllers available from a default fitbenchmarking install
import inspect
import numpy as np
import os
from unittest import TestCase
import pytest
from pytest import test_type as TEST_TYPE

from fitbenchmarking import mock_problems
from fitbenchmarking.controllers.base_controller import Controller
from fitbenchmarking.controllers.bumps_controller import BumpsController
from fitbenchmarking.controllers.controller_factory import ControllerFactory
from fitbenchmarking.controllers.dfo_controller import DFOController
from fitbenchmarking.controllers.minuit_controller import MinuitController
from fitbenchmarking.controllers.scipy_controller import ScipyController
from fitbenchmarking.controllers.scipy_ls_controller import ScipyLSController

if TEST_TYPE != "default":
    from fitbenchmarking.controllers.gsl_controller import GSLController
    from fitbenchmarking.controllers.mantid_controller import MantidController
    from fitbenchmarking.controllers.ralfit_controller import RALFitController


from fitbenchmarking.parsing.parser_factory import parse_problem_file
from fitbenchmarking.utils import exceptions
from fitbenchmarking.utils.options import Options
from fitbenchmarking.jacobian.SciPyFD_2point_jacobian import ScipyTwoPoint


def make_fitting_problem(file_name='cubic.dat'):
    """
    Helper function that returns a simple fitting problem
    """
    options = Options()

    bench_prob_dir = os.path.dirname(inspect.getfile(mock_problems))
    fname = os.path.join(bench_prob_dir, file_name)

    fitting_problem = parse_problem_file(fname, options)
    fitting_problem.correct_data()
    jac = ScipyTwoPoint(fitting_problem)
    fitting_problem.jac = jac
    return fitting_problem


class DummyController(Controller):
    """
    Minimal instantiatable subclass of Controller class for testing
    """

    def setup(self):
        self.setup_result = 53

    def fit(self):
        raise NotImplementedError

    def cleanup(self):
        raise NotImplementedError

    def error_flags(self):
        raise NotImplementedError

class ControllerSharedTesting:

    def controller_run_test(self, controller):
        """
        Utility function to run controller and check output is in generic form

        :param controller: Controller to test, with setup already completed
        :type controller: Object derived from BaseSoftwareController
        """
        controller.parameter_set = 0
        controller.prepare()
        controller.fit()
        controller.cleanup()

        assert len(controller.final_params) == len(controller.initial_params)

    def check_converged(self, controller):
        """
        Utility function to check controller.cleanup() produces a success flag

        :param controller: Controller to test, with setup already completed
        :type controller: Object derived from BaseSoftwareController
        """
        controller.cleanup()
        assert controller.flag == 0

    def check_max_iterations(self, controller):
        """
        Utility function to check controller.cleanup() produces a maximum
        iteration flag

        :param controller: Controller to test, with setup already completed
        :type controller: Object derived from BaseSoftwareController
        """
        controller.cleanup()
        assert controller.flag == 1

    def check_diverged(self, controller):
        """
        Utility function to check controller.cleanup() produces a fail

        :param controller: Controller to test, with setup already completed
        :type controller: Object derived from BaseSoftwareController
        """
        controller.cleanup()
        assert controller.flag == 2


class BaseControllerTests(TestCase):
    """
    Tests for base software controller class methods.
    """

    def setUp(self):
        self.problem = make_fitting_problem()

    def test_data(self):
        """
        BaseSoftwareController: Test data is read into controller correctly
        """

        controller = DummyController(self.problem)

        if self.problem.start_x is not None:
            assert min(controller.data_x) >= self.problem.start_x
        if self.problem.end_x is not None:
            assert max(controller.data_x) <= self.problem.end_x

        assert len(controller.data_e) == len(controller.data_x)
        assert len(controller.data_e) == len(controller.data_y)

        self.assertTrue(all(x in self.problem.data_x
                            for x in controller.data_x))
        self.assertTrue(all(y in self.problem.data_y
                            for y in controller.data_y))

        e_is_default = self.problem.data_e is None
        if not e_is_default:
            self.assertTrue(all(e in self.problem.data_e
                                for e in controller.data_e))

    def test_prepare(self):
        """
        BaseSoftwareController: Test prepare function
        """
        controller = DummyController(self.problem)
        controller.minimizer = 'test'
        controller.parameter_set = 0
        controller.prepare()
        assert controller.setup_result == 53

    def test_eval_chisq_no_errors(self):
        """
        BaseSoftwareController: Test eval_chisq function
        """
        controller = DummyController(self.problem)

        params = np.array([1, 2, 3, 4])
        x = np.array([6, 2, 32, 4])
        y = np.array([1, 21, 3, 4])
        e = None

        result = self.problem.eval_r_norm(params=params, x=x, y=y, e=e)

        assert controller.eval_chisq(params=params, x=x, y=y, e=e) == result

    def test_eval_chisq_with_errors(self):
        """
        BaseSoftwareController: Test eval_chisq function
        """
        controller = DummyController(self.problem)

        params = np.array([1, 2, 3, 4])
        x = np.array([6, 2, 32, 4])
        y = np.array([1, 21, 3, 4])
        e = np.array([.5, .003, 1, 2])

        result = self.problem.eval_r_norm(params=params, x=x, y=y, e=e)

        assert controller.eval_chisq(params=params, x=x, y=y, e=e) == result

    def test_check_flag_attr(self):
        """
        BaseSoftwareController: Test check_attributes function for _flag
                                attribute
        """
        controller = DummyController(self.problem)
        with self.assertRaises(exceptions.ControllerAttributeError):
            controller.check_attributes()
        controller.flag = 1
        controller.final_params = [1]
        controller.check_attributes()

    def test_check_valid_flag(self):
        """
        BaseSoftwareController: Test flag setting with valid values
        """
        controller = DummyController(self.problem)
        controller.final_params = [1]

        for flag in [0, 1, 2, 3]:
            controller.flag = flag

    def test_check_invalid_flag(self):
        """
        BaseSoftwareController: Test flag setting with invalid values
        """
        controller = DummyController(self.problem)
        controller.final_params = [1, 2, 3, 4, 5]
        with self.assertRaises(exceptions.ControllerAttributeError):
            controller.flag = 10

    def test_check_final_params_attr(self):
        """
        BaseSoftwareController: Test check_attributes function for final_params
                                attribute
        """
        controller = DummyController(self.problem)
        controller.final_params = [1, 2, 3, 4, 5]
        controller.flag = 3
        controller.check_attributes()

    def test_check_invalid_final_params(self):
        """
        BaseSoftwareController: Test final_params setting with invalid values
        """
        controller = DummyController(self.problem)
        controller.flag = 1
        controller.final_params = [1, np.inf]
        with self.assertRaises(exceptions.ControllerAttributeError):
            controller.check_attributes()

    def test_validate_minimizer_true(self):
        controller = DummyController(self.problem)
        controller.algorithm_check = {'all': ['min1', 'min2']}
        algorithm_type = 'all'
        minimizer = 'min1'
        controller.validate_minimizer(minimizer, algorithm_type)

    def test_validate_minimizer_false(self):
        controller = DummyController(self.problem)
        controller.algorithm_check = {'all': ['min1', 'min2']}
        algorithm_type = 'all'
        minimizer = 'min_unknown'
        with self.assertRaises(exceptions.UnknownMinimizerError):
            controller.validate_minimizer(minimizer, algorithm_type)


class DefaultControllerTests(TestCase):
    """
    Tests for each controller class
    """

    def setUp(self):
        self.problem = make_fitting_problem()
        self.shared_tests = ControllerSharedTesting()

    def test_bumps(self):
        """
        BumpsController: Test for output shape
        """
        controller = BumpsController(self.problem)
        controller.minimizer = 'amoeba'
        self.shared_tests.controller_run_test(controller)

        controller._status = 0
        self.shared_tests.check_converged(controller)
        controller._status = 2
        self.shared_tests.check_max_iterations(controller)
        controller._status = 1
        self.shared_tests.check_diverged(controller)

    def test_dfo(self):
        """
        DFOController: Tests for output shape
        """
        controller = DFOController(self.problem)
        # test one from each class
        minimizers = ['dfogn',
                      'dfols']
        for minimizer in minimizers:
            controller.minimizer = minimizer
            self.shared_tests.controller_run_test(controller)

            controller._status = 0
            self.shared_tests.check_converged(controller)
            controller._status = 2
            self.shared_tests.check_max_iterations(controller)
            controller._status = 5
            self.shared_tests.check_diverged(controller)

    def test_minuit(self):
        """
        MinuitController: Tests for output shape
        """
        controller = MinuitController(self.problem)
        controller.minimizer = 'minuit'
        self.shared_tests.controller_run_test(controller)
        controller._status = 0
        self.shared_tests.check_converged(controller)
        controller._status = 2
        self.shared_tests.check_diverged(controller)

    def test_scipy(self):
        """
        ScipyController: Test for output shape
        """
        controller = ScipyController(self.problem)
        controller.minimizer = 'CG'
        self.shared_tests.controller_run_test(controller)

        controller._status = 0
        self.shared_tests.check_converged(controller)
        controller._status = 2
        self.shared_tests.check_max_iterations(controller)
        controller._status = 1
        self.shared_tests.check_diverged(controller)

    def test_scipy_ls(self):
        """
        ScipyLSController: Test for output shape
        """
        controller = ScipyLSController(self.problem)
        controller.minimizer = 'lm'

        self.shared_tests.controller_run_test(controller)

        controller._status = 1
        self.shared_tests.check_converged(controller)
        controller._status = 0
        self.shared_tests.check_max_iterations(controller)
        controller._status = -1
        self.shared_tests.check_diverged(controller)


@pytest.mark.skipif("TEST_TYPE == 'default'")
class ExternalControllerTests(TestCase):
    """
    Tests for each controller class
    """

    def setUp(self):
        self.problem = make_fitting_problem()
        self.shared_tests = ControllerSharedTesting()

    def test_mantid(self):
        """
        MantidController: Test for output shape
        """
        controller = MantidController(self.problem)
        controller.minimizer = 'Levenberg-Marquardt'
        self.shared_tests.controller_run_test(controller)

        controller._status = "success"
        self.shared_tests.check_converged(controller)
        controller._status = "Failed to converge"
        self.shared_tests.check_max_iterations(controller)
        controller._status = "Failed"
        self.shared_tests.check_diverged(controller)

    def test_mantid_multifit(self):
        """
        MantidController: Additional bespoke test for multifit
        """
        file_path = os.path.join('multifit_set', 'multifit.txt')
        problem = make_fitting_problem(file_path)

        controller = MantidController(problem)
        controller.minimizer = 'Levenberg-Marquardt'

        controller.parameter_set = 0
        controller.prepare()
        controller.fit()
        controller.cleanup()

        self.assertEqual(len(controller.final_params), len(controller.data_x),
                         'Multifit did not return a result for each data file')

        self.assertEqual(len(controller.final_params[0]),
                         len(controller.initial_params),
                         'Incorrect number of final params.')

    def test_mantid_singlefit_chisquared(self):
        """
        Test the override in Mantid conroller is working correctly for
        evaluating chi_squared (SingleFit).
        """
        m_controller = MantidController(self.problem)
        b_controller = DummyController(self.problem)
        params = np.array([1, 2, 3, 4])
        x = np.array([6, 2, 32, 4])
        y = np.array([1, 21, 3, 4])
        e = np.array([.5, .003, 1, 2])

        expected = b_controller.eval_chisq(params=params, x=x, y=y, e=e)
        actual = m_controller.eval_chisq(params=params, x=x, y=y, e=e)

        self.assertEqual(expected, actual,
                         'Mantid controller found a different chi squared'
                         ' for single fit problem.')

    def test_mantid_multifit_chisquared(self):
        """
        Test the override in Mantid conroller is working correctly for
        evaluating chi_squared (MultiFit).
        """
        m_controller = MantidController(self.problem)
        b_controller = DummyController(self.problem)
        params = [np.array([1, 2, 3, 4]),
                  np.array([1, 2, 3, 4]),
                  np.array([1, 2, 3, 4])]
        xs = [np.array([6, 2, 32, 4]),
              np.array([6, 2, 32, 4]),
              np.array([6, 2, 32, 4])]
        ys = [np.array([1, 21, 3, 4]),
              np.array([1, 21, 3, 4]),
              np.array([1, 21, 3, 4])]
        es = [np.array([.5, .003, 1, 2]),
              np.array([.5, .003, 1, 2]),
              np.array([.5, .003, 1, 2])]

        expected = [b_controller.eval_chisq(params=p, x=x, y=y, e=e)
                    for x, y, e, p in zip(xs, ys, es, params)]
        actual = m_controller.eval_chisq(params=params, x=xs, y=ys, e=es)

        self.assertListEqual(
            expected, actual,
            'Mantid controller found a different chi squared for multi fit'
            ' problem.')

    def test_gsl(self):
        """
        GSLController: Tests for output shape
        """
        controller = GSLController(self.problem)
        # test one from each class
        minimizers = ['lmsder',
                      'nmsimplex',
                      'conjugate_pr']
        for minimizer in minimizers:
            controller.minimizer = minimizer
            self.shared_tests.controller_run_test(controller)

            controller.flag = 0
            self.shared_tests.check_converged(controller)
            controller.flag = 1
            self.shared_tests.check_max_iterations(controller)
            controller.flag = 2
            self.shared_tests.check_diverged(controller)

    def test_ralfit(self):
        """
        RALFitController: Tests for output shape
        """
        controller = RALFitController(self.problem)
        minimizers = ['gn', 'gn_reg', 'hybrid', 'hybrid_reg']
        for minimizer in minimizers:
            controller.minimizer = minimizer
            self.shared_tests.controller_run_test(controller)

            controller._status = 0
            self.shared_tests.check_converged(controller)
            controller._status = 2
            self.shared_tests.check_diverged(controller)


class FactoryTests(TestCase):
    """
    Tests for the ControllerFactory
    """

    def test_default_imports(self):
        """
        Test that the factory returns the correct default class for inputs
        """
        valid = ['scipy_ls', 'bumps']
        valid_names = ['scipyls', 'bumps']
        invalid = ['foo', 'bar', 'hello', 'r2d2']
        self.check_valid(valid, valid_names)
        self.check_invalid(invalid)

    @pytest.mark.skipif("TEST_TYPE == 'default'")
    def test_external_imports(self):
        """
        Test that the factory returns the correct external class for inputs
        """
        valid = ['mantid', 'ralfit']
        valid_names = ['mantid', 'ralfit']
        self.check_valid(valid, valid_names)

    def check_valid(self, valid, valid_names):

        for software, v in zip(valid, valid_names):
            controller = ControllerFactory.create_controller(software)
            self.assertTrue(controller.__name__.lower().startswith(v))

    def check_invalid(self, invalid):

        for software in invalid:
            self.assertRaises(exceptions.NoControllerError,
                              ControllerFactory.create_controller,
                              software)



"""
Implements a controller for DFO-GN
http://people.maths.ox.ac.uk/robertsl/dfogn/
"""

import dfogn
import dfols
import numpy as np

from fitbenchmarking.controllers.base_controller import Controller


class DFOController(Controller):
    """
    Controller for the DFO-{GN/LS} fitting software.
    """

    def __init__(self, problem):
        """
        Initialises variables used for temporary storage.

        :param problem: Problem to fit
        :type problem: FittingProblem
        """
        super(DFOController, self).__init__(problem)

        self._soln = None
        self._popt = None
        self._pinit = None
        self.algorithm_check = {
            'all': ['dfogn', 'dfols'],
            'ls': ['dfogn', 'dfols'],
            'deriv_free': ['dfogn', 'dfols'],
            'general': [None]}

    def setup(self):
        """
        Setup for DFO
        """
        self._pinit = np.asarray(self.initial_params)

    def fit(self):
        """
        Run problem with DFO.
        """
        if self.minimizer == 'dfogn':
            self._soln = dfogn.solve(self.problem.eval_r,
                                     self._pinit)
        elif self.minimizer == 'dfols':
            self._soln = dfols.solve(self.problem.eval_r,
                                     self._pinit)

        self._popt = self._soln.x
        self._status = self._soln.flag

    def cleanup(self):
        """
        Convert the result to a numpy array and populate the variables results
        will be read from.
        """
        if self._status == 0:
            self.flag = 0
        elif self._status == 2:
            self.flag = 1
        else:
            self.flag = 2

        self.final_params = self._popt



"""
FitBenchmarking results object
"""

from __future__ import (absolute_import, division, print_function)

# To store the results in the object requires more than the default
# max arguments and sanitised_name setter requires no use of self
# pylint: disable=too-many-arguments, no-self-use


class FittingResult:
    """
    Minimal definition of a class to hold results from a
    fitting problem test.
    """

    def __init__(self, options, problem, jac, initial_params, params,
                 name=None, chi_sq=None, runtime=None,
                 minimizer=None, error_flag=None, dataset_id=None):
        """
        Initialise the Fitting Result

        :param options: Options used in fitting
        :type options: utils.options.Options
        :param problem: The Problem definition for the fit
        :type problem: parsing.fitting_problem.FittingProblem
        :param jac: The Jacobian definition
        :type jac: fitbenchmarking.jacobian.base_controller.Jacobian subclass
        :param initial_params: The starting parameters for the fit
        :type initial_params: list of float
        :param params: The parameters found by the fit
        :type params: list of float or list of list of float
        :param name: Name of the result, defaults to None
        :type name: str, optional
        :param chi_sq: The score for the fitting, defaults to None
        :type chi_sq: float or list of float, optional
        :param runtime: The average runtime of the fit, defaults to None
        :type runtime: float or list of float, optional
        :param minimizer: The name of the minimizer used, defaults to None
        :type minimizer: str, optional
        :param error_flag: [description], defaults to None
        :type error_flag: [type], optional
        :param dataset_id: The index of the dataset (Only used for MultiFit),
                           defaults to None
        :type dataset_id: int, optional
        """
        self.options = options
        self.problem = problem
        self.jac = jac
        self.name = name if name is not None else \
            problem.name

        self.chi_sq = chi_sq
        if dataset_id is None:
            self.data_x = problem.data_x
            self.data_y = problem.data_y
            self.data_e = problem.data_e
            self.sorted_index = problem.sorted_index

            self.params = params
            self.chi_sq = chi_sq

        else:
            self.data_x = problem.data_x[dataset_id]
            self.data_y = problem.data_y[dataset_id]
            self.data_e = problem.data_e[dataset_id]
            self.sorted_index = problem.sorted_index[dataset_id]

            self.params = params[dataset_id]
            self.chi_sq = chi_sq[dataset_id]

        self.runtime = runtime

        self.min_chi_sq = None
        self.min_runtime = None

        # Minimizer for a certain problem and its function definition
        self.minimizer = minimizer

        # String interpretations of the params
        self.ini_function_params = self.problem.get_function_params(
            params=initial_params)
        self.fin_function_params = self.problem.get_function_params(
            params=self.params)

        # Controller error handling
        self.error_flag = error_flag

        # Paths to various output files
        self.support_page_link = ''
        self.start_figure_link = ''
        self.figure_link = ''

        # Error written to support page if plotting failed
        # Default can be overwritten with more information
        self.figure_error = 'Plotting Failed'

        self._norm_acc = None
        self._norm_runtime = None
        self.is_best_fit = False

    @property
    def norm_acc(self):
        """
        Getting function for norm_acc attribute

        :return: normalised accuracy value
        :rtype: float
        """
        if self._norm_acc is None:
            self._norm_acc = self.chi_sq / self.min_chi_sq
        return self._norm_acc

    @norm_acc.setter
    def norm_acc(self, value):
        """
        Stores the normalised accuracy and updates the value

        :param value: New value for norm_runtime
        :type value: float
        """
        self._norm_acc = value

    @property
    def norm_runtime(self):
        """
        Getting function for norm_runtime attribute

        :return: normalised runtime value
        :rtype: float
        """
        if self._norm_runtime is None:
            self._norm_runtime = self.runtime / self.min_runtime
        return self._norm_runtime

    @norm_runtime.setter
    def norm_runtime(self, value):
        """
        Stores the normalised runtime and updates the value

        :param value: New value for norm_runtime
        :type value: float
        """
        self._norm_runtime = value

    @property
    def sanitised_name(self):
        """
        Sanitise the problem name into one which can be used as a filename.

        :return: sanitised name
        :rtype: str
        """
        return self.name.replace(',', '').replace(' ', '_')
    
    @property
    def sanitised_min_name(self):
        """
        Sanitise the minimizer name into one which can be used as a filename.

        :return: sanitised name
        :rtype: str
        """
        return self.minimizer.replace(':', '').replace(' ', '_')

    @sanitised_name.setter
    def sanitised_name(self, value):
        raise RuntimeError('sanitised_name can not be edited')
    
    @sanitised_min_name.setter
    def sanitised_min_name(self, value):
        raise RuntimeError('sanitised_min_name can not be edited')



"""
Tests for fitbenchmarking.core.fitting_benchmarking.loop_over_fitting_software
"""
from __future__ import (absolute_import, division, print_function)
import inspect
import os
import unittest
import numpy as np

from fitbenchmarking import mock_problems
from fitbenchmarking.utils import fitbm_result, output_grabber
from fitbenchmarking.core.fitting_benchmarking import \
    loop_over_fitting_software
from fitbenchmarking.parsing.parser_factory import parse_problem_file
from fitbenchmarking.utils.options import Options
from fitbenchmarking.jacobian.SciPyFD_2point_jacobian import ScipyTwoPoint
from fitbenchmarking.utils.exceptions import UnsupportedMinimizerError

# Defines the module which we mock out certain function calls for
FITTING_DIR = "fitbenchmarking.core.fitting_benchmarking"

# Due to structure of tests, some variables may not be previously defined
# in the init function
# pylint: disable=attribute-defined-outside-init
def make_fitting_problem(file_name='cubic.dat', minimizers=None):
    """
    Helper function that returns a simple fitting problem
    """
    options = Options()
    if minimizers:
        options.minimizers = minimizers

    bench_prob_dir = os.path.dirname(inspect.getfile(mock_problems))
    fname = os.path.join(bench_prob_dir, file_name)

    fitting_problem = parse_problem_file(fname, options)
    fitting_problem.correct_data()
    jac = ScipyTwoPoint(fitting_problem)
    fitting_problem.jac = jac
    return fitting_problem


class LoopOverSoftwareTests(unittest.TestCase):
    """
    loop_over_fitting_software tests
    """

    def setUp(self):
        """
        Setting up problem for tests
        """
        self.problem = make_fitting_problem()
        self.options = self.problem.options
        self.options.software = ["scipy", "dfo"]
        self.minimizers = self.options.minimizers
        self.grabbed_output = output_grabber.OutputGrabber(self.options)
        self.start_values_index = 0
        self.scipy_len = len(self.options.minimizers["scipy"])
        self.dfo_len = len(self.options.minimizers["dfo"])
        self.result_args = {'options': self.options,
                            'problem': self.problem,
                            'jac': self.problem.jac,
                            'initial_params': self.problem.starting_values[0],
                            'params': [],
                            'chi_sq': 1}

    def mock_func_call(self, *args, **kwargs):
        """
        Mock function to be used instead of loop_over_minimizers
        """
        minimizer_failed = self.minimizer_failed[self.count]
        results_problem = self.results_problem[self.count]
        self.count += 1
        return results_problem, minimizer_failed

    def shared_test(self, expected_list_len, expected_problem_fails,
                    expected_minimizer_failed):
        """
        Shared tests for the `loop_over_fitting_software` function

        :param expected_list_len: number of expect fitting results
        :type expected_list_len: int
        :param expected_problem_fails: expected list of failed problems
        :type expected_problem_fails: list
        :param expected_minimizer_failed: expected list of failed minimizers
        :type expected_minimizer_failed: list
        """
        results, problem_fails, unselected_minimzers = \
            loop_over_fitting_software(self.problem,
                                       self.options,
                                       self.start_values_index,
                                       self.grabbed_output)
        assert len(results) == expected_list_len
        assert problem_fails == expected_problem_fails
        for i, (keys, values) in enumerate(unselected_minimzers.items()):
            assert keys in self.options.software
            assert values == expected_minimizer_failed[i]

    @unittest.mock.patch('{}.loop_over_minimizers'.format(FITTING_DIR))
    def test_run_software(self, loop_over_minimizers):
        """
        Checks that results are produced for all minimizers within the
        softwares
        """
        self.count = 0
        self.minimizer_failed = [[], []]
        self.problem_fails = []
        self.results_problem = \
            [[fitbm_result.FittingResult(**self.result_args)
              for i in range(self.scipy_len)],
             [fitbm_result.FittingResult(**self.result_args)
              for i in range(self.dfo_len)]]
        loop_over_minimizers.side_effect = self.mock_func_call
        expected_list_len = self.scipy_len + self.dfo_len
        expected_problem_fails = self.problem_fails
        expected_minimizer_failed = self.minimizer_failed
        self.shared_test(expected_list_len, expected_problem_fails,
                         expected_minimizer_failed)

    @unittest.mock.patch('{}.loop_over_minimizers'.format(FITTING_DIR))
    def test_run_software_failed_minimizers(self, loop_over_minimizers):
        """
        Checks that the failed minimizers are reported
        """
        self.count = 0
        self.minimizer_failed = [['Powell'], ['dfogn', 'dfols']]
        failed_scipy = 1
        failed_dfo = 2
        self.problem_fails = []
        self.results_problem = \
            [[fitbm_result.FittingResult(**self.result_args)
              for i in range(self.scipy_len - failed_scipy)],
             [fitbm_result.FittingResult(**self.result_args)
              for i in range(self.dfo_len - failed_dfo)]]
        loop_over_minimizers.side_effect = self.mock_func_call
        expected_list_len = self.scipy_len + self.dfo_len - \
            failed_scipy - failed_dfo
        expected_problem_fails = self.problem_fails
        expected_minimizer_failed = self.minimizer_failed
        self.shared_test(expected_list_len, expected_problem_fails,
                         expected_minimizer_failed)

    @unittest.mock.patch('{}.loop_over_minimizers'.format(FITTING_DIR))
    def test_run_software_all_failed_minimizers(self, loop_over_minimizers):
        """
        Tests that when all minimizers raise and exception for a problem
        that this is reported correctly
        """
        self.count = 0
        self.minimizer_failed = [self.options.minimizers[s]
                                 for s in self.options.software]
        self.result_args['chi_sq'] = np.inf
        self.problem_fails = ['cubic']
        self.results_problem = [[fitbm_result.FittingResult(**self.result_args)
                                 for i in range(self.scipy_len)],
                                [fitbm_result.FittingResult(**self.result_args)
                                 for i in range(self.dfo_len)]]
        loop_over_minimizers.side_effect = self.mock_func_call
        expected_list_len = 0
        expected_problem_fails = self.problem_fails
        expected_minimizer_failed = self.minimizer_failed
        self.shared_test(expected_list_len, expected_problem_fails,
                         expected_minimizer_failed)

    def test_incorrect_software(self):
        """
        Tests an exception is raised when an incorrect software is selected
        """
        self.options.software = ['incorrect_software']
        with self.assertRaises(UnsupportedMinimizerError):
            _, _, _ = \
                loop_over_fitting_software(self.problem,
                                           self.options,
                                           self.start_values_index,
                                           self.grabbed_output)


if __name__ == "__main__":
    unittest.main()



"""
Higher level functions that are used for plotting the fit plot and a starting
guess plot.
"""
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import os


class Plot(object):
    """
    Class providing plotting functionality.
    """

    def __init__(self, best_result, options, figures_dir):
        self.problem = best_result.problem
        self.options = options

        self.result = best_result

        self.figures_dir = figures_dir

        self.legend_location = "upper left"
        self.title_size = 10

        # These define the styles of the 4 types of plot
        self.data_plot_options = {"label": "Data",
                                  "zorder": 0,
                                  "color": "black",
                                  "marker": "x",
                                  "linestyle": '',
                                  "linewidth": 1}
        self.ini_guess_plot_options = {"label": "Starting Guess",
                                       "zorder": 1,
                                       "color": "#ff6699",
                                       "marker": "",
                                       "linestyle": '-',
                                       "linewidth": 3}
        self.best_fit_plot_options = {"zorder": 3,
                                      "color": '#6699ff',
                                      "marker": "",
                                      "linestyle": ':',
                                      "linewidth": 3}
        self.fit_plot_options = {"zorder": 2,
                                 "color": "#99ff66",
                                 "marker": "",
                                 "linestyle": '-',
                                 "linewidth": 3}

        # Create a single reusable plot containing the problem data.
        # We store a line here, which is updated to change the graph where we
        # know the rest of the graph is untouched between plots.
        # This is more efficient that the alternative of creating a new graph
        # every time.
        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(1, 1, 1)
        self.line_plot = None
        # Plot the data that functions were fitted to
        self.plot_data(self.options.use_errors,
                       self.data_plot_options)
        # reset line_plot as base data won't need updating
        self.line_plot = None

        # Store sorted x values for plotting
        self.x = self.result.data_x[self.result.sorted_index]

    def __del__(self):
        """
        Close the matplotlib figure
        """
        plt.close(self.fig)

    def format_plot(self):
        """
        Performs post plot processing to annotate the plot correctly
        """
        self.ax.set_xlabel(r"Time ($\mu s$)")
        self.ax.set_ylabel("Arbitrary units")
        self.ax.set_title(self.result.name,
                          fontsize=self.title_size)
        self.ax.legend(loc=self.legend_location)
        self.fig.set_tight_layout(True)

    def plot_data(self, errors, plot_options, x=None, y=None):
        """
        Plots the data given

        :param errors: whether fit minimizer uses errors
        :type errors: bool
        :param plot_options: Values for style of the data to plot,
                                 for example color and zorder
        :type plot_options: dict
        :param x: x values to be plotted
        :type x: np.array
        :param y: y values to be plotted
        :type y: np.array
        """
        if x is None:
            x = self.result.data_x
        if y is None:
            y = self.result.data_y
        if errors:
            # Plot with errors
            self.ax.clear()
            self.ax.errorbar(x, y, yerr=self.result.data_e,
                             **plot_options)
        else:
            # Plot without errors
            if self.line_plot is None:
                # Create a new line and store
                self.line_plot = self.ax.plot(x, y, **plot_options)[0]
            else:
                # Update line instead of recreating
                self.line_plot.set_data(x, y)
                # Update style
                for k, v in plot_options.items():
                    try:
                        getattr(self.line_plot, 'set_{}'.format(k))(v)
                    except AttributeError:
                        pass

                self.fig.canvas.draw()

    def plot_initial_guess(self):
        """
        Plots the initial guess along with the data and stores in a file

        :return: path to the saved file
        :rtype: str
        """
        # Parse which starting values to use from result name
        start_index = self.result.name.partition('Start')[2].split(',')[0]
        if start_index:
            start_index = int(start_index.strip())
        else:
            start_index = 1

        ini_guess = self.problem.starting_values[start_index - 1].values()
        self.plot_data(errors=False,
                       plot_options=self.ini_guess_plot_options,
                       x=self.x,
                       y=self.problem.eval_f(list(ini_guess), self.x))
        self.format_plot()
        file = "start_for_{0}.png".format(self.result.sanitised_name)
        file_name = os.path.join(self.figures_dir, file)
        self.fig.savefig(file_name)
        return file

    def plot_best(self, minimizer, params):
        """
        Plots the fit along with the data using the "best_fit" style
        and saves to a file

        :param minimizer: name of the best fit minimizer
        :type minimizer: str
        :param params: fit parameters returned from the best fit minimizer
        :type params: list

        :return: path to the saved file
        :rtype: str
        """
        label = 'Best Fit ({})'.format(minimizer)

        # Plot line and save.
        # This should have the style of fit plot options and the colour of
        # best fit plot options.
        # Then the plot should be saved and updated to use only best fit style
        # for future plots.
        plot_options_dict = self.fit_plot_options.copy()
        plot_options_dict['label'] = label
        plot_options_dict['color'] = self.best_fit_plot_options['color']

        y = self.problem.eval_f(params, self.x)
        self.plot_data(errors=False,
                       plot_options=plot_options_dict,
                       y=y)
        self.format_plot()
        file = "{}_fit_for_{}.png".format(self.result.sanitised_min_name,
                                          self.result.sanitised_name)
        file_name = os.path.join(self.figures_dir, file)
        self.fig.savefig(file_name)

        # Update to correct linestyle
        plot_options_dict = self.best_fit_plot_options.copy()
        plot_options_dict['label'] = label
        self.plot_data(errors=False,
                       plot_options=plot_options_dict,
                       x=self.x,
                       y=y)

        # Make sure line wont be replaced by resetting line_plot
        self.line_plot = None
        return file

    def plot_fit(self, minimizer, params):
        """
        Updates self.line to show the fit using the passed in params.
        If self.line is empty it will create a new line.
        Stores the plot in a file

        :param minimizer: name of the fit minimizer
        :type minimizer: str
        :param params: fit parameters returned from the best fit minimizer
        :type params: list

        :return: path to the saved file
        :rtype: str
        """
        plot_options_dict = self.fit_plot_options.copy()
        plot_options_dict['label'] = minimizer

        self.plot_data(errors=False,
                       plot_options=plot_options_dict,
                       x=self.x,
                       y=self.problem.eval_f(params, self.x))
        self.format_plot()
        file = "{}_fit_for_{}.png".format(self.result.sanitised_min_name,
                                          self.result.sanitised_name)
        file_name = os.path.join(self.figures_dir, file)
        self.fig.savefig(file_name)
        return file



"""
Implements a controller for GSL
https://www.gnu.org/software/gsl/
using the pyGSL python interface
https://sourceforge.net/projects/pygsl/
"""
import numpy as np

from pygsl import multifit_nlin, multiminimize, errno
from pygsl import _numobj as numx

from scipy.optimize._numdiff import approx_derivative

from fitbenchmarking.controllers.base_controller import Controller
from fitbenchmarking.utils.exceptions import UnknownMinimizerError


class GSLController(Controller):
    """
    Controller for the GSL fitting software
    """

    def __init__(self, problem):
        """
        Initializes variable used for temporary storage

        :param problem: Problem to fit
        :type problem: FittingProblem
        """
        super(GSLController, self).__init__(problem)

        self._solver = None
        self._residual_methods = None
        self._function_methods_no_jac = None
        self._function_methods_with_jac = None
        self._gradient_tol = None
        self._abserror = None
        self._relerror = None
        self._maxits = None
        self.algorithm_check = {
            'all': ['lmsder', 'lmder', 'nmsimplex', 'nmsimplex2',
                    'conjugate_pr', 'conjugate_fr', 'vector_bfgs',
                    'vector_bfgs2', 'steepest_descent'],
            'ls': ['lmsder', 'lmder'],
            'deriv_free': ['nmsimplex', 'nmsimplex2'],
            'general': ['nmsimplex', 'nmsimplex2', 'conjugate_pr',
                        'conjugate_fr', 'vector_bfgs', 'vector_bfgs2',
                        'steepest_descent']}

    def _prediction_error(self, p, data=None):
        """
        Utility function to call problem.eval_r with correct args

        :param p: parameters
        :type p: list
        :param data: x data, this is discarded as the defaults can be used.
        :type data: N/A
        :return: result from problem.eval_r
        :rtype: numpy array
        """
        return self.problem.eval_r(p)

    def _jac(self, p, data=None):
        """
        Utility function to call jac.eval with correct args

        :param p: parameters
        :type p: list
        :param data: x data, this is discarded as the defaults can be used.
        :type data: N/A
        :return: result from jac.eval
        :rtype: numpy array
        """
        return self.problem.jac.eval(p)

    def _fdf(self, p, data=None):
        """
        Utility function to return results from eval_r and eval_j as a tuple.

        :param p: parameters
        :type p: list
        :param data: x data, this is discarded as the defaults can be used.
        :type data: N/A
        :return: result from problem.eval_r and eval_j
        :rtype: (numpy array, numpy array)
        """
        f = self.problem.eval_r(p)
        df = self.problem.jac.eval(p)
        return f, df

    def _chi_squared(self, p, data=None):
        """
        Utility function to call problem.eval_r_norm with correct args

        :param p: parameters
        :type p: list
        :param data: x data, this is discarded as the defaults can be used.
        :type data: N/A
        :return: result from problem.eval_r_norm
        :rtype: numpy array
        """
        return self.problem.eval_r_norm(p)

    def _jac_chi_squared(self, p, data=None):
        """
        Utility function to get jacobian for problem.eval_r_norm

        :param p: parameters
        :type p: list
        :param data: x data, this is discarded as the defaults can be used.
        :type data: N/A
        :return: jacobian approximation for problem.eval_r_norm
        :rtype: numpy array
        """
        j = approx_derivative(self.problem.eval_r_norm, p)
        return j

    def _chi_squared_fdf(self, p, data=None):
        """
        Utility function to return results from eval_r_norm and
        _jac_chi_squared as a tuple.

        :param p: parameters
        :type p: list
        :param data: x data, this is discarded as the defaults can be used.
        :type data: N/A
        :return: result from problem.eval_r_norm and _jac_chi_squared
        :rtype: (numpy array, numpy array)
        """
        f = self.problem.eval_r_norm(p)
        df = self._jac_chi_squared(p)
        return f, df

    def setup(self):
        """
        Setup for GSL
        """
        data = numx.array([self.data_x,
                           self.data_y,
                           self.data_e])
        n = len(self.data_x)
        p = len(self.initial_params)
        pinit = numx.array(self.initial_params)

        self._residual_methods = ['lmsder',
                                  'lmder']
        self._function_methods_no_jac = ['nmsimplex',
                                         'nmsimplex2']
        self._function_methods_with_jac = ['conjugate_pr',
                                           'conjugate_fr',
                                           'vector_bfgs',
                                           'vector_bfgs2',
                                           'steepest_descent']

        # set up the system
        if self.minimizer in self._residual_methods:
            mysys = multifit_nlin.gsl_multifit_function_fdf(
                self._prediction_error,
                self._jac,
                self._fdf,
                data,
                n,
                p)
            self._solver = getattr(multifit_nlin, self.minimizer)(mysys, n, p)
        elif self.minimizer in self._function_methods_no_jac:
            mysys = multiminimize.gsl_multimin_function(self._chi_squared,
                                                        data,
                                                        p)
            self._solver = getattr(multiminimize, self.minimizer)(mysys, p)
        elif self.minimizer in self._function_methods_with_jac:
            mysys = multiminimize.gsl_multimin_function_fdf(
                self._chi_squared,
                self._jac_chi_squared,
                self._chi_squared_fdf,
                data,
                p)
            self._solver = getattr(multiminimize, self.minimizer)(mysys, p)
        else:
            raise UnknownMinimizerError(
                "No {} minimizer for GSL".format(self.minimizer))

        # Set up initialization parameters
        #
        # These have been chosen to be consistent with the parameters
        # used in Mantid.
        initial_steps = 1.0 * numx.array(np.ones(p))
        step_size = 0.1
        tol = 1e-4
        self._gradient_tol = 1e-3
        self._abserror = 1e-4
        self._relerror = 1e-4
        self._maxits = 500

        if self.minimizer in self._residual_methods:
            self._solver.set(pinit)
        elif self.minimizer in self._function_methods_no_jac:
            self._solver.set(pinit, initial_steps)
        else:
            self._solver.set(pinit, step_size, tol)

    def fit(self):
        """
        Run problem with GSL
        """
        for _ in range(self._maxits):
            status = self._solver.iterate()
            # check if the method has converged
            if self.minimizer in self._residual_methods:
                x = self._solver.getx()
                dx = self._solver.getdx()
                status = multifit_nlin.test_delta(dx, x,
                                                  self._abserror,
                                                  self._relerror)
            elif self.minimizer in self._function_methods_no_jac:
                simplex_size = self._solver.size()
                status = multiminimize.test_size(simplex_size,
                                                 self._abserror)
            else:  # must be in function_methods_with_jac
                gradient = self._solver.gradient()
                status = multiminimize.test_gradient(gradient,
                                                     self._gradient_tol)
            if status == errno.GSL_SUCCESS:
                self.flag = 0
                break
            elif status != errno.GSL_CONTINUE:
                self.flag = 2
        else:
            self.flag = 1

    def cleanup(self):
        """
        Convert the result to a numpy array and populate the variables results
        will be read from
        """
        self.final_params = self._solver.getx()



"""
Utility functions for creating/deleting directories.
"""

import os
import shutil


def results(results_dir):
    """
    Creates the results folder in the working directory.

    :param results_dir: path to the results directory, results dir name
    :type results_dir: str

    :return: proper path to the results directory
    :rtype: str
    """

    if not isinstance(results_dir, str):
        raise TypeError("results_dir must be a string!")

    if not os.path.exists(results_dir):
        os.makedirs(results_dir)

    return results_dir


def group_results(results_dir, group_name):
    """
    Creates the results directory for a specific group.
    e.g. fitbenchmarking/results/Neutron/

    :param results_dir: path to directory that holds all the results
    :type results_dir: str
    :param group_name: name of the problem group
    :type group_name: str

    :return: path to folder group specific results dir
    :rtype: str
    """

    if isinstance(group_name, str):
        group_dir = os.path.join(results_dir, group_name)
    else:
        raise TypeError('Type of variable group_name is required '
                        'to be a string, type(group_name) '
                        '= {}'.format(type(group_name)))
    if os.path.exists(group_dir):
        shutil.rmtree(group_dir)
    os.makedirs(group_dir)
    return group_dir


def support_pages(group_results_dir):
    """
    Creates the support_pages directory in the group results directory.

    :param group_results_dir: path to the group results directory
    :type group_results_dir: str

    :return: path to the figures directory
    :rtype: str
    """
    support_pages_dir = os.path.join(group_results_dir, "support_pages")
    if not os.path.exists(support_pages_dir):
        os.makedirs(support_pages_dir)

    return support_pages_dir


def figures(support_pages_dir):
    """
    Creates the figures directory inside the support_pages directory.

    :param support_pages_dir: path to the support pages directory
    :type support_pages_dir: str

    :return: path to the figures directory
    :rtype: str
    """
    figures_dir = os.path.join(support_pages_dir, "figures")
    if not os.path.exists(figures_dir):
        os.makedirs(figures_dir)

    return figures_dir

def css(results_dir):
    """
    Creates a local css directory inside the results directory.

    :param support_pages_dir: path to the results directory
    :type support_pages_dir: str

    :return: path to the local css directory
    :rtype: str
    """
    local_css_dir = os.path.join(results_dir, "css")
    if not os.path.exists(local_css_dir):
        os.makedirs(local_css_dir)

    return local_css_dir


def del_contents_of_dir(directory):
    """
    Delete contents of a directory, including other directories.

    :param directory: the target directory
    :type directory: str
    """

    for file in os.listdir(directory):
        file_path = os.path.join(directory, file)
        if os.path.isfile(file_path):
            os.unlink(file_path)
        elif os.path.isdir(file_path):
            shutil.rmtree(file_path)



"""
Table tests
"""

from __future__ import (absolute_import, division, print_function)
import unittest
from inspect import getfile
import os
import shutil
import numpy as np

import fitbenchmarking
from fitbenchmarking.results_processing.tables import SORTED_TABLE_NAMES
from fitbenchmarking.results_processing.tables import generate_table
from fitbenchmarking.results_processing.tables import create_results_tables
from fitbenchmarking.jacobian.SciPyFD_2point_jacobian import ScipyTwoPoint
from fitbenchmarking.core.results_output import preproccess_data
from fitbenchmarking.parsing.fitting_problem import FittingProblem
from fitbenchmarking.utils.fitbm_result import FittingResult
from fitbenchmarking.utils.options import Options


# By design both fitting_function_1 and fitting_function_2 need data as an
# argument
# pylint: disable=unused-argument
def fitting_function_1(data, x1, x2):
    """
    Fitting function evaluator

    :param data: x data
    :type data: numpy array
    :param x1: fitting parameter
    :type x1: float
    :param x2: fitting parameter
    :type x2: float

    :return: y data values evaluated from the function of the problem
    :rtype: numpy array
    """
    return x1 * np.sin(x2)


def fitting_function_2(data, x1, x2):
    """
    Fitting function evaluator

    :param data: x data
    :type data: numpy array
    :param x1: fitting parameter
    :type x1: float
    :param x2: fitting parameter
    :type x2: float

    :return: y data values evaluated from the function of the problem
    :rtype: numpy array
    """
    return x1 * x2

# pylint: enable=unused-argument


def generate_mock_results():
    """
    Generates results to test against

    :return: best results calculated using the chi_sq value, list of results
             and the options
    :rtype: tuple(list of best results,
                  list of list fitting results,
                  Options object)
    """
    software = 'scipy_ls'
    options = Options()
    options.software = [software]
    num_min = len(options.minimizers[options.software[0]])
    data_x = np.array([[1, 4, 5], [2, 1, 5]])
    data_y = np.array([[1, 2, 1], [2, 2, 2]])
    data_e = np.array([[1, 1, 1], [1, 2, 1]])
    func = [fitting_function_1, fitting_function_2]
    problems = [FittingProblem(options), FittingProblem(options)]

    params_in = [[[.3, .11], [.04, 2], [3, 1], [5, 0]],
                 [[4, 2], [3, .006], [.3, 10], [9, 0]]]

    starting_values = [{"a": .3, "b": .11}, {"a": 0, "b": 0}]
    error_in = [[1, 0, 2, 0],
                [0, 1, 3, 1]]
    link_in = [['link1', 'link2', 'link3', 'link4'],
               ['link5', 'link6', 'link7', 'link8']]
    acc_in = [[1, 5, 2, 1.54],
              [7, 3, 5, 1]]
    runtime_in = [[1e-2, 2.2e-3, 4.2e-5, 9.8e-1],
                  [3.0e-10, 5.0e-14, 1e-7, 4.3e-12]]

    results_out = []
    for i, p in enumerate(problems):
        p.data_x = data_x[i]
        p.data_y = data_y[i]
        p.data_e = data_e[i]
        p.function = func[i]
        p.name = "prob_{}".format(i)
        results = []
        for j in range(num_min):
            p.starting_values = starting_values
            jac = ScipyTwoPoint(p)
            r = FittingResult(options=options, problem=p, jac=jac,
                              initial_params=starting_values,
                              params=params_in[i][j])
            r.chi_sq = acc_in[i][j]
            r.runtime = runtime_in[i][j]
            r.error_flag = error_in[i][j]
            r.support_page_link = link_in[i][j]
            r.minimizer = options.minimizers[software][j]
            results.append(r)
        results_out.append(results)
    best = preproccess_data(results_out)
    return best, results_out, options


class GenerateTableTests(unittest.TestCase):
    """
    Class that tests the generate_table function within
    fitbenchmarking.results_processing.tables
    """

    def setUp(self):
        """
        Setup up method for test
        """
        self.best, self.results, self.options = generate_mock_results()
        root = os.path.dirname(getfile(fitbenchmarking))

        self.expected_results_dir = os.path.join(root, 'results_processing',
                                                 'tests', 'expected_results')

    def test_tables_correct(self):
        """
        Test that the tables are equal to the expected output stored in
        fitbenchmarking/results_processing/tests/expected_results
        """
        for suffix in SORTED_TABLE_NAMES:
            _, html_table, txt_table = generate_table(
                self.results, self.best, self.options, "group_dir",
                ["pp_1", "pp_2"], "table_name", suffix)
            html_table_name = os.path.join(self.expected_results_dir,
                                           "{}.html".format(suffix))
            txt_table_name = os.path.join(self.expected_results_dir,
                                          "{}.txt".format(suffix))

            for f, t in zip([html_table_name, txt_table_name],
                            [html_table, txt_table]):
                self.compare_files(f, t)

    def compare_files(self, expected_table, table):
        """
        Compares two tables line by line

        :param expected_table: imported html output from expected results in
                               fitbenchmarking/results_processing/tests/
                               expected_results
        :type expected_table: str
        :param table: table generated using generate_table in
                      fitbenchmarking.results_processing.tables
        :type table: str
        """
        with open(expected_table, 'r') as f:
            expected = f.readlines()

        file_extension = expected_table.split('.')[1]
        if file_extension == 'txt':
            html_id_expected = ''
            html_id = ''
        elif file_extension == 'html':
            html_id_expected = expected[1].strip(' ').split('row')[0][1:]
            html_id = table.splitlines()[1].strip(' ').split('row')[0][1:]
        diff = []
        for act_line, exp_line in zip(table.splitlines(), expected):
            exp_line = '' if exp_line is None else exp_line.strip('\n')
            act_line = '' if act_line is None else act_line.strip('\n')
            exp_line = exp_line.replace(html_id_expected, html_id)
            # to pass on windows need to first do this before comparing
            act_line = act_line.replace('href=\"..\\', 'href=\"../')
            if act_line != exp_line:
                diff.append([exp_line, act_line])
        self.assertListEqual([], diff)


class CreateResultsTableTests(unittest.TestCase):
    """
    Class that tests the generate_table function within
    fitbenchmarking.results_processing.create_results_tables
    """

    def setUp(self):
        """
        Setup up method for test
        """
        self.best, self.results, self.options = generate_mock_results()
        root = os.path.dirname(getfile(fitbenchmarking))

        self.group_dir = os.path.join(root, 'results_processing',
                                      'tests', 'results')
        os.mkdir(self.group_dir)
        self.group_name = 'test_name'

    def tearDown(self):
        """
        Deletes temporary folder and results produced
        """
        if os.path.exists(self.group_dir):
            shutil.rmtree(self.group_dir)

    def test_generate_table_page(self):
        """
        Checks to see whether files with the correct name are produced.
        """
        create_results_tables(self.options, self.results, self.best,
                              self.group_name, self.group_dir,
                              ["pp_1", "pp_2"], [], {'min1': []})
        for suffix in SORTED_TABLE_NAMES:

            for table_type in ['html', 'txt']:
                table_name = '{}_{}_weighted_table.{}'.format(self.group_name,
                                                              suffix,
                                                              table_type)
                file_name = os.path.join(self.group_dir, table_name)
                self.assertTrue(os.path.isfile(file_name))


if __name__ == "__main__":
    unittest.main()



"""
Implements a controller for the scipy fitting software.
In particular, here for the scipy minimize solver for general minimization problems
"""

from scipy.optimize import minimize

from fitbenchmarking.controllers.base_controller import Controller


class ScipyController(Controller):
    """
    Controller for the Scipy fitting software.
    """

    def __init__(self, problem):
        """
        Initialises variable used for temporary storage.
        """
        super(ScipyController, self).__init__(problem)

        self._popt = None
        self.algorithm_check = {
            'all': ['Nelder-Mead', 'Powell', 'CG', 'BFGS', 'Newton-CG',
                    'L-BFGS-B', 'TNC', 'SLSQP'],
            'ls': [None],
            'deriv_free': ['Nelder-Mead', 'Powell'],
            'general': ['Nelder-Mead', 'Powell', 'CG', 'BFGS',
                        'Newton-CG', 'L-BFGS-B', 'TNC', 'SLSQP']}

    def setup(self):
        """
        Setup problem ready to be run with SciPy
        """
        if self.minimizer == "lm-scipy":
            self.minimizer = "lm"

        self.options = {'maxiter': 500}

    def eval_jac(self, x, *args):
        """
        Wrapper for problem.eval_j to form the approximate Jacobian for
        problem.eval_r_norm

        :param x: The parameter values to find the Jacobian at
        :type x: list

        :return: Approximation of the Jacobian
        :rtype: numpy array
        """
        out = self.problem.jac.eval(params=x,
                                    func=self.problem.eval_r_norm,
                                    *args)
        return out

    def fit(self):
        """
        Run problem with Scipy.
        """
        # Neither the Nelder-Mead or Powell minimizers require a Jacobian
        # so are run without that argument.
        if self.minimizer == "Nelder-Mead" or self.minimizer == "Powell":
            self.result = minimize(fun=self.problem.eval_r_norm,
                                   x0=self.initial_params,
                                   method=self.minimizer,
                                   options=self.options)
        else:
            self.result = minimize(fun=self.problem.eval_r_norm,
                                   x0=self.initial_params,
                                   method=self.minimizer,
                                   jac=self.eval_jac,
                                   options=self.options)
        self._popt = self.result.x
        self._status = self.result.status

    def cleanup(self):
        """
        Convert the result to a numpy array and populate the variables results
        will be read from.
        """
        if self._status == 0:
            self.flag = 0
        elif self._status == 2:
            self.flag = 1
        else:
            self.flag = 2

        self.final_params = self._popt



"""
Implements a controller for RALFit
https://github.com/ralna/RALFit
"""

import ral_nlls

from fitbenchmarking.controllers.base_controller import Controller
from fitbenchmarking.utils.exceptions import UnknownMinimizerError


class RALFitController(Controller):
    """
    Controller for the RALFit fitting software.
    """

    def __init__(self, problem):
        """
        Initialises variable used for temporary storage.

        :param problem: Problem to fit
        :type problem: FittingProblem
        """
        super(RALFitController, self).__init__(problem)

        self._popt = None
        self._options = {}
        self.algorithm_check = {
            'all': ['gn', 'hybrid', 'gn_reg', 'hybrid_reg'],
            'ls': ['gn', 'hybrid', 'gn_reg', 'hybrid_reg'],
            'deriv_free': [None],
            'general': [None]}

    def setup(self):
        """
        Setup for RALFit
        """

        # Use bytestrings explicitly as thats whats expected in RALFit and
        # python 3 defaults to unicode.
        self._options[b"maxit"] = 500
        if self.minimizer == "gn":
            self._options[b"model"] = 1
            self._options[b"nlls_method"] = 4
        elif self.minimizer == "gn_reg":
            self._options[b"model"] = 1
            self._options[b"type_of_method"] = 2
        elif self.minimizer == "hybrid":
            self._options[b"model"] = 3
            self._options[b"nlls_method"] = 4
        elif self.minimizer == "hybrid_reg":
            self._options[b"model"] = 3
            self._options[b"type_of_method"] = 2
        else:
            raise UnknownMinimizerError(
                "No {} minimizer for RALFit".format(self.minimizer))

    def fit(self):
        """
        Run problem with RALFit.
        """
        self._popt = ral_nlls.solve(self.initial_params,
                                    self.problem.eval_r,
                                    self.problem.jac.eval,
                                    options=self._options)[0]
        self._status = 0 if self._popt is not None else 1

    def cleanup(self):
        """
        Convert the result to a numpy array and populate the variables results
        will be read from.
        """
        if self._status == 0:
            self.flag = 0
        else:
            self.flag = 2

        self.final_params = self._popt



"""
compare table
"""
import numpy as np
import os
from fitbenchmarking.results_processing.base_table import Table

GRAD_TOL = 1e-1
RES_TOL = 1e-8


class LocalMinTable(Table):
    """

    The local min results shows a ``True`` or ``False`` value together with
    :math:`\\frac{|| J^T r||}{||r||}`. The ``True`` or ``False`` indicates
    whether the software finds a minimum with respect to the following
    criteria:


    - :math:`||r|| \\leq` RES\\_TOL,
    - :math:`|| J^T r|| \\leq` GRAD\\_TOL,
    - :math:`\\frac{|| J^T r||}{||r||} \\leq` GRAD\\_TOL,

    where :math:`J` and :math:`r` are the Jacobian and residual of
    :math:`f(x, p)`, respectively. The tolerances can be found in the results
    object.

    """

    def __init__(self, results, best_results, options, group_dir,
                 pp_locations, table_name):
        """
        Initialise the local minimizer table which shows given the
        conditioners stated in the doc string whether the final parameters
        are a local minimum

        :param results: results nested array of objects
        :type results: list of list of
                       fitbenchmarking.utils.fitbm_result.FittingResult
        :param best_results: best result for each problem
        :type best_results: list of
                        fitbenchmarking.utils.fitbm_result.FittingResult
        :param options: Options used in fitting
        :type options: utils.options.Options
        :param group_dir: path to the directory where group results should be
                          stored
        :type group_dir: str
        :param pp_locations: tuple containing the locations of the
                             performance profiles (acc then runtime)
        :type pp_locations: tuple(str,str)
        :param table_name: Name of the table
        :type table_name: str
        """
        self.name = 'local_min'
        super(LocalMinTable, self).__init__(results, best_results, options,
                                            group_dir, pp_locations,
                                            table_name)

        self.has_pp = True
        self.pp_filenames = \
                        [os.path.relpath(pp,group_dir) for pp in pp_locations]

    def get_values(self, results_dict):
        """
        Gets the main values to be reported in the tables

        :param results_dict: dictionary containing results where the keys
                             are the problem sets and the values are lists
                             of results objects
        :type results_dict: dictionary

        :return: a dictionary containing true or false values whether the
                 return parameters is a local minimizer and a dictionary
                 containing :math:`\\frac{|| J^T r||}{||r||}` values
        :rtype: tuple(dict, dict)
        """
        local_min = {}
        norm_rel = {}
        for key, value in results_dict.items():
            local_min[key] = []
            norm_rel[key] = []
            for i, v in enumerate(value):
                if v.params is None:
                    norm_rel[key].append(np.inf)
                    local_min[key].append("False")
                else:
                    res = v.problem.eval_r(
                        v.params, v.data_x, v.data_y, v.data_e)
                    jac = v.jac.eval(v.params, v.problem.eval_r,
                                     x=v.data_x, y=v.data_y, e=v.data_e)
                    min_test = np.matmul(res, jac)

                    norm_r = np.linalg.norm(res)
                    norm_min_test = np.linalg.norm(min_test)

                    norm_rel[key].append(norm_min_test / norm_r)

                    for r, m, n in zip([norm_r],
                                       [norm_min_test],
                                       [norm_rel[key][i]]):
                        if r <= RES_TOL or m <= GRAD_TOL or n <= GRAD_TOL:
                            local_min[key].append("True")
                        else:
                            local_min[key].append("False")

        return local_min, norm_rel

    def get_colour(self, results):
        """
        Uses the local minimizer dictionary values to set the HTML colour

        :param results: a dictionary containing true or false values whether
                        the return parameters is a local minimizer and a
                        dictionary containing
                        :math:`\\frac{|| J^T r||}{||r||}` values
        :type results: tuple

        :return: dictionary containing error codes from the minimizers
        :rtype: dict
        """
        local_min, _ = results
        colour = {key: [self.html_colours[0]
                        if v == "True" else self.html_colours[-1]
                        for v in value]
                  for key, value in local_min.items()}
        return colour

    def display_str(self, results):
        """
        Function that combines the True and False value from variable local_min
        with the normalised residual

        :param results: a dictionary containing true or false values whether
                        the return parameters is a local minimizer and a
                        dictionary containing
                        :math:`\\frac{|| J^T r||}{||r||}` values
        :type results: tuple

        :return: dictionary containing the string representation of the values
                 in the table.
        :rtype: dict
        """
        local_min, norm_rel = results
        template = self.output_string_type['abs']
        table_output = {}
        for key in local_min.keys():
            table_output[key] = [a + " (" + template.format(r) + ")"
                                 for a, r in zip(local_min[key],
                                                 norm_rel[key])]
        return table_output



"""
Tests for FitBenchmarking object
"""
from __future__ import (absolute_import, division, print_function)
import inspect
import os
import unittest
import numpy as np

from fitbenchmarking import mock_problems
from fitbenchmarking.parsing.parser_factory import parse_problem_file
from fitbenchmarking.jacobian.SciPyFD_2point_jacobian import ScipyTwoPoint
from fitbenchmarking.utils.fitbm_result import FittingResult
from fitbenchmarking.utils.options import Options


class FitbmResultTests(unittest.TestCase):
    """
    Tests for FitBenchmarking results object
    """

    def setUp(self):
        """
        Setting up FitBenchmarking results object
        """
        self.options = Options()
        mock_problems_dir = os.path.dirname(inspect.getfile(mock_problems))
        problem_dir = os.path.join(mock_problems_dir, "cubic.dat")
        self.problem = parse_problem_file(problem_dir, self.options)
        self.problem.correct_data()

        self.chi_sq = 10
        self.minimizer = "test_minimizer"
        self.runtime = 0.01
        self.params = np.array([1, 3, 4, 4])
        self.initial_params = np.array([0, 0, 0, 0])
        self.jac = ScipyTwoPoint(self.problem)
        self.result = FittingResult(
            options=self.options, problem=self.problem, jac=self.jac,
            chi_sq=self.chi_sq, runtime=self.runtime, minimizer=self.minimizer,
            initial_params=self.initial_params, params=self.params,
            error_flag=0)

        self.min_chi_sq = 0.1
        self.result.min_chi_sq = self.min_chi_sq
        self.min_runtime = 1
        self.result.min_runtime = self.min_runtime

    def test_init_with_dataset_id(self):
        """
        Tests to check that the multifit id is setup correctly
        """
        chi_sq = [10, 5, 1]
        minimizer = "test_minimizer"
        runtime = 0.01
        params = [np.array([1, 3, 4, 4]),
                  np.array([2, 3, 57, 8]),
                  np.array([4, 2, 5, 1])]
        initial_params = np.array([0, 0, 0, 0])

        self.problem.data_x = [np.array([3, 2, 1, 4]),
                               np.array([5, 1, 2, 3]),
                               np.array([6, 7, 8, 1])]
        self.problem.data_y = [np.array([2, 1, 7, 40]),
                               np.array([8, 9, 4, 2]),
                               np.array([7, 4, 4, 2])]
        self.problem.data_e = [np.array([1, 1, 1, 1]),
                               np.array([2, 2, 2, 1]),
                               np.array([2, 3, 4, 4])]
        self.problem.sorted_index = [np.array([2, 1, 0, 3]),
                                     np.array([1, 2, 3, 0]),
                                     np.array([3, 0, 1, 2])]
        jac = ScipyTwoPoint(self.problem)
        self.result = FittingResult(
            options=self.options, problem=self.problem, jac=jac,
            chi_sq=chi_sq, runtime=runtime, minimizer=minimizer,
            initial_params=initial_params, params=params,
            error_flag=0, dataset_id=1)

        self.assertTrue(
            np.isclose(self.result.data_x, self.problem.data_x[1]).all())
        self.assertTrue(
            np.isclose(self.result.data_y, self.problem.data_y[1]).all())
        self.assertTrue(
            np.isclose(self.result.data_e, self.problem.data_e[1]).all())
        self.assertTrue(
            np.isclose(self.result.sorted_index,
                       self.problem.sorted_index[1]).all())

        self.assertTrue(np.isclose(params[1], self.result.params).all())
        self.assertEqual(chi_sq[1], self.result.chi_sq)

    def test_norm_acc(self):
        """
        Test that sanitised names are correct.
        """
        expected = self.chi_sq / self.min_chi_sq
        self.assertEqual(self.result.norm_acc, expected)

    def test_norm_runtime(self):
        """
        Test that sanitised names are correct.
        """
        expected = self.runtime / self.min_runtime
        self.assertEqual(self.result.norm_runtime, expected)

    def test_sanitised_name(self):
        """
        Test that sanitised names are correct.
        """
        self.result.name = 'test, name with commas'
        self.assertEqual(self.result.sanitised_name, 'test_name_with_commas')

    def test_sanitised_min_name(self):
        """
        Test that sanitised minimizer names are correct.
        """
        self.result.minimizer = 'test: name with colon'
        self.assertEqual(self.result.sanitised_min_name, 'test_name_with_colon')


if __name__ == "__main__":
    unittest.main()



"""
Tests for fitbenchmarking.core.fitting_benchmarking.benchmark
"""
from __future__ import (absolute_import, division, print_function)
import inspect
import copy
import os
import unittest
from unittest import mock

from fitbenchmarking import mock_problems
from fitbenchmarking.utils import fitbm_result
from fitbenchmarking.core.fitting_benchmarking import benchmark
from fitbenchmarking.parsing.parser_factory import parse_problem_file
from fitbenchmarking.utils.options import Options
from fitbenchmarking.jacobian.SciPyFD_2point_jacobian import ScipyTwoPoint
from fitbenchmarking.utils.exceptions import NoResultsError

# Defines the module which we mock out certain function calls for
FITTING_DIR = "fitbenchmarking.core.fitting_benchmarking"


def make_fitting_problem(file_name='cubic.dat', minimizers=None):
    """
    Helper function that returns a simple fitting problem
    """
    options = Options()
    if minimizers:
        options.minimizers = minimizers

    bench_prob_dir = os.path.dirname(inspect.getfile(mock_problems))
    fname = os.path.join(bench_prob_dir, file_name)

    fitting_problem = parse_problem_file(fname, options)
    fitting_problem.correct_data()
    jac = ScipyTwoPoint(fitting_problem)
    fitting_problem.jac = jac
    return fitting_problem


def dict_test(expected, actual):
    """
    Test to check two dictionaries are the same

    :param expected: expected dictionary result
    :type expected: dict
    :param actual: actual dictionary result
    :type actual: dict
    """
    for key in actual.keys():
        assert key in expected.keys()
        assert sorted(actual[key]) == sorted(expected[key])


class BenchmarkTests(unittest.TestCase):
    """
    benchmark tests
    """

    def setUp(self):
        """
        Setting up problem for tests
        """
        self.problem = make_fitting_problem()
        self.options = Options()
        self.options.software = ["scipy"]
        self.scipy_len = len(self.options.minimizers["scipy"])
        bench_prob_dir = os.path.dirname(inspect.getfile(mock_problems))
        self.default_parsers_dir = os.path.join(bench_prob_dir,
                                                "default_parsers")
        self.all_minimzers = copy.copy(self.options.minimizers)

    def shared_tests(self, expected_names, expected_unselected_minimzers,
                     expected_minimzers):
        """
        Shared tests for the `benchmark` function. The function test

        :param expected_names: expected sorted list of problem names
        :type expected_names: list
        :param expected_unselected_minimzers: expected unselected minimizers
        :type expected_unselected_minimzers: dict
        :param expected_minimzers: expected minimizers
        :type expected_minimzers: dict
        """
        results, failed_problems, unselected_minimzers = \
            benchmark(self.options, self.default_parsers_dir)

        assert len(results) == len(expected_names)
        for i, name in enumerate(expected_names):
            assert all(p.name == name for p in results[i])

        assert failed_problems == []
        dict_test(expected_unselected_minimzers, unselected_minimzers)
        dict_test(expected_minimzers, self.options.minimizers)

    @mock.patch('{}.loop_over_benchmark_problems'.format(FITTING_DIR))
    def test_check_no_unselected_minimizers(self,
                                            loop_over_benchmark_problems):
        """
        Checks benchmarking runs with no unselected minimizers
        """
        # define mock return for loop_over_benchmark_problems
        problem_names = ["random_1", "random_3", "random_2"]
        results = []
        for name in problem_names:
            result_args = {'options': self.options,
                           'problem': self.problem,
                           'jac': self.problem.jac,
                           'initial_params': self.problem.starting_values[0],
                           'params': [],
                           'chi_sq': 1,
                           'name': name}
            list_results = [fitbm_result.FittingResult(**result_args)
                            for j in range(self.scipy_len)]
            results.extend(list_results)
        problem_fails = []
        expected_unselected_minimzers = {"scipy": []}
        loop_over_benchmark_problems.return_value = \
            (results, problem_fails, expected_unselected_minimzers)

        # run shared test and see if it match expected
        expected_problem_names = sorted(problem_names)
        expected_minimzers = copy.copy(self.all_minimzers)
        self.shared_tests(expected_problem_names,
                          expected_unselected_minimzers,
                          expected_minimzers)

    @mock.patch('{}.loop_over_benchmark_problems'.format(FITTING_DIR))
    def test_check_unselected_minimizers(self, loop_over_benchmark_problems):
        """
        Checks benchmarking runs with a few unselected minimizers
        """
        # define mock return for loop_over_benchmark_problems
        problem_names = ["random_1", "random_3", "random_2"]
        expected_names = sorted(problem_names)
        results = []
        for name in problem_names:
            result_args = {'options': self.options,
                           'problem': self.problem,
                           'jac': self.problem.jac,
                           'initial_params': self.problem.starting_values[0],
                           'params': [],
                           'chi_sq': 1,
                           'name': name}
            list_results = [fitbm_result.FittingResult(**result_args)
                            for j in range(self.scipy_len)]
            results.extend(list_results)
        problem_fails = []
        expected_unselected_minimzers = {"scipy": ['SLSQP', 'Powell', 'CG']}
        loop_over_benchmark_problems.return_value = \
            (results, problem_fails, expected_unselected_minimzers)

        # run shared test and see if it match expected
        expected_minimzers = copy.copy(self.all_minimzers)
        for keys, minimzers in expected_unselected_minimzers.items():
            diff = set(expected_minimzers[keys]) - set(minimzers)
            expected_minimzers[keys] = [x for x in expected_minimzers[keys]
                                        if x in diff]

        self.shared_tests(expected_names, expected_unselected_minimzers,
                          expected_minimzers)


if __name__ == "__main__":
    unittest.main()
